#include <iostream>
// #include <cassert>
#include <utility>
#include <algorithm>
#include <cmath>
#include "source/network.h"
#include "source/arrivalchain.h"
#include "source/departurechain.h"
#include "source/queues.h"
#include "source/general.h"
#include "source/lane.h"
#include "source/abpmap.h"
#include "source/deparrqueues.h"


extern int TOTAL_QUEUE_COPIES;
extern int TOTAL_QUEUE_ASSIGNMENTS;
extern int TOTAL_QUEUE_DEFAULT_CONSTRUCTIONS;
extern int TOTAL_QUEUE_NORMAL_CONSTRUCTIONS;
extern int TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS;
extern int TOTAL_QUEUE_MIX_CONSTRUCTIONS;
extern int TOTAL_QUEUE_DEPARTURE_CONSTRUCTORS;
extern int TOTAL_EMPLACE;

extern int DELETED_QUEUES;


std::vector<int> getTimes(int duration, int offset, int cycle){
    std::vector<int> result;
    for (int i = offset; i < duration; i += cycle){
        result.push_back(i);
    }
    return result;
}


double run(int duration){
    double total_queue = -1;
    try{
        std::string input_file_name = std::string("grid_3x3_2") + "_";
        Network network(input_file_name + "lanes.txt", input_file_name + "arrival_rates.txt",
                    input_file_name + "routing.txt", input_file_name + "green_schedule.txt", input_file_name + "state_step_0.txt");
        total_queue = network.run(duration);         std::cout << std::fixed << "\nTotal queue length: " << total_queue << " m * sec\n";
    } catch (char const* exception) {
        switchStyle("red");
        std::cout << exception << "\n";
        switchStyle("reset");
        std::cout << "Beee!" << std::endl;
    }
    return total_queue;
}




void printConstructorsStatistics(){
    std::cout << "Queue constructors (default, simple, normal, mixes, copies, departures, assignments (emplaces)):\n"
              << TOTAL_QUEUE_DEFAULT_CONSTRUCTIONS << " " << TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS << " " << TOTAL_QUEUE_NORMAL_CONSTRUCTIONS << " "
              << TOTAL_QUEUE_MIX_CONSTRUCTIONS << " "
              << TOTAL_QUEUE_COPIES << " " << TOTAL_QUEUE_DEPARTURE_CONSTRUCTORS << " " << TOTAL_QUEUE_ASSIGNMENTS << " (" << TOTAL_EMPLACE << ")\n";
    std::cout << "Together without default: " << TOTAL_QUEUE_COPIES + TOTAL_EMPLACE + TOTAL_QUEUE_ASSIGNMENTS + TOTAL_QUEUE_DEPARTURE_CONSTRUCTORS
                 + TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS + TOTAL_QUEUE_NORMAL_CONSTRUCTIONS + TOTAL_QUEUE_MIX_CONSTRUCTIONS << "\n";
    std::cout << "(Explicitely) deleted queues: " << DELETED_QUEUES << "\n";
}


int main(int argc, char **argv) {
    int duration = DURATION;
    if (argc > 1) {
        std::cout << argc << "\n";
        std::cout << argv[0] << "\n";
        throw "Incorrect input: no parameters are required\n";
    }

    run(duration);     
    // printConstructorsStatistics();
    return 0;
}
