#ifndef ARRIVALCHAIN_H
#define ARRIVALCHAIN_H
#include <vector>
#include <map>
#include "state.h"
#include "general.h"
// #include "queues.h"

class Queues;

class ArrivalChain{
    std::vector<std::map<char, ArrivalState>> chain;
    int last_fully_known;
    
    void addStateProbability(int second, char a_key, double value);
public:
//     ArrivalChain(int duration);  // an arrival chain for an internal lane
    ArrivalChain(int duration, int travel_time);  // an arrival chain for an internal lane
    ArrivalChain(int duration, int travel_time, std::vector<int> & arrivals, int starting_second);  // an arrival chain for an internal lane
    explicit ArrivalChain(std::vector<double>& arrival_rates); // an arrival chain for an external lane
    ArrivalChain(std::vector<double>& arrival_rates, std::vector<int> & arrivals, int starting_second); // an arrival chain for an external lane
    void printStates (int second) const;
    void printStatesProbabilities (int second);
    std::map<char, double> getNormalisedTransitions(int second, char key);
    std::map<char, std::map<char, double>> getNormalisedTransitions(int second);
    double getTransitionFlow(int second, char key_from, char key_to) {return chain[second][key_from].getTransition(key_to);}
    void addFlow(int second, char a_key_from, char a_key_to, double flow);
    std::map<char, std::map<char, double>> getNormalisedTransitionsToKeys(int second);
    std::map<char, std::map<char, double>> getNormalisedTransitionsToKeys(int second, int downstream_second, Queues & queues);
    std::map<char, std::map<char, double>> getNormalisedFullTransitionsToKeys(int second, Queues & queues);
    std::map<char, ArrivalState>& getStates(int second);
    char getFirstState(int second);
    bool hasState(int second, char a_key) const;
    std::map<KeyIA, double> getUnknownTransitionProbabilities(int second);
    double getUnknownTransitionProbability(int second, char a_key);
    double getUntransitionedProbability(int second, char a_key);
    std::map<KeyIA, double> getBlockingProbabilities(int downstream_second, Queues & queues);
    
    double getStateProbability(int second, char a_key) const;
    int getSize() { return chain.size(); }
    bool checkTotalProbability(int second);
    bool checkTotalTransitionProbability(int second);
    bool checkTransitionedFlows(int second);
    
    void addDefaultTransitions(int travel_time_second, double unblocked);
    void addTransitionedProbabilities(int second, MapCharChar & transitioned, MapCharChar & transitioned_full, double unblocked);
};

#endif // ARRIVALCHAIN_H
