#include "deparrqueues.h"

int DELETED_QUEUES = 0;

void ArrQueues::addQueueOrInsert(int ind, Queue & queue_to_add){
    if (hasKey(ind)){
        queues[ind].addQueue(queue_to_add);
    } else {
        queues.insertItem(ind, queue_to_add);
    }
}

void ArrQueues::insert(int ind, const Queue& queue){
    queues.insertItem(ind, queue);
}

void ArrQueues::insert(char a_key, const Queue& queue){
    insert(getInd(a_key), queue);
}




void DepArrQueues::addQueues(DepArrQueues& to_add){
    Pair d_key;
    for (auto & d_key_queues: to_add.queues){
        d_key = d_key_queues.first;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind)){
                queues[d_key].addQueueOrInsert(a_ind, d_key_queues.second[a_ind]);
            }
        }
    }
}


void DepArrQueues::removeDownstreamParts(int second, int downstream_second){
    for (auto d_key_queues = queues.begin(); d_key_queues != queues.end(); /* no increment*/){
        bool no_keys_left = true;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues->second.hasKey(a_ind)){
                if (d_key_queues->second[a_ind].removeDownstreamPart(second, downstream_second)){ // the queue emptied
                    d_key_queues->second.eraseItem(a_ind);
                } else {
                    no_keys_left = false;
                }
            }
        }
        if (no_keys_left){
            // remove this d_key
            d_key_queues = queues.erase(d_key_queues);
        } else {
            ++d_key_queues;
        }
    }
}

void DepArrQueues::removeDownstreamParts(int m_second){
    for (auto d_key_queues = queues.begin(); d_key_queues != queues.end(); /* no increment*/){
        bool no_keys_left = true;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues->second.hasKey(a_ind)){
                if (d_key_queues->second[a_ind].removeDownstreamPart(m_second)){ // the queue emptied
                    d_key_queues->second.eraseItem(a_ind);
                } else {
                    no_keys_left = false;
                }
            }
        }
        if (no_keys_left){
            // remove this d_key
            d_key_queues = queues.erase(d_key_queues);
        } else {
            ++d_key_queues;
        }
    }
}

double DepArrQueues::getTotalProbability() const{
    double total = 0;
//     std::map<char, double> per_a_key;
    for (auto & d_key_queues: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind)){
                total += d_key_queues.second[a_ind].getTotalProbability();
//                 addValue(per_a_key, getAKey(a_ind), d_key_queues.second[a_ind].getTotalProbability());
            }
        }
    }
//     std::cout << "per a_key " << per_a_key << "\n";
    return total;
}


void DepArrQueues::renormaliseQueues(double total){
    for (auto & d_key_queues: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind)){
                d_key_queues.second[a_ind].renormaliseQueue(total);
            }
        }
    }
}

void DepArrQueues::printQueues() const{
    std::cout << "printing queues\n";
    for (auto & d_key_queues: queues){
        std::cout << d_key_queues.first << ":\n";
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_queues.second.hasKey(a_ind)){
                std::cout << a_ind << ":\n";
                d_key_queues.second[a_ind].printQueue();
            }
        }
    }
    std::cout << "finish printing queues\n";
}

void DepArrQueues::deleteEmptyQueues(){
//     Pair d_key;
    for (auto d_key_iter = queues.begin(); d_key_iter != queues.end(); /* no increment*/){
//         d_key = d_key_iter->first;
        int a_keys_number = 0;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter->second.hasKey(a_ind)){ 
                if (d_key_iter->second[a_ind].hasAlmostZeroProbability()){ // delete queue
                    d_key_iter->second.eraseItem(a_ind);
                    ++DELETED_QUEUES;
                } else {
                    ++a_keys_number;
                }
            }                    
        }
        if (a_keys_number == 0) // delete d_key
            queues.erase(d_key_iter++);
        else  // leave d_key
            ++d_key_iter;
    }
}

double DepArrQueues::getAverageQueue(){
    double result = 0;
    for (auto & d_key_iter: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind))
                result += d_key_iter.second[a_ind].getExpectedQueueLength();
        }
    }
    return result;
}


double DepArrQueues::getAverageSquareQueue(int start_from /* = 0*/){
    double result = 0;
    for (auto & d_key_iter: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind))
                result += d_key_iter.second[a_ind].getExpectedQueueSquare(start_from);
        }
    }
    return result;
}


PairMapChar DepArrQueues::getProbabilitiesPerAKeys(){
    PairMapChar result; // first MapChar correspond to not full probabilities and second MapChar to full probabilities
    double total, full;
    for (auto & d_key_iter: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind)){
                total = d_key_iter.second[a_ind].getTotalProbability();
                full = d_key_iter.second[a_ind].getAlmostFullProbability(a_ind == KEY_A);
                addValue(result.first, getAKey(a_ind), total - full);
                addValue(result.second, getAKey(a_ind), full);
            }
        }
    }
    return result;
}


void DepArrQueues::eraseItem(Pair d_key){
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        if (queues[d_key].hasKey(a_ind)){
            queues[d_key].eraseItem(a_ind);
        }
    }
    queues.erase(d_key);
}


int DepArrQueues::getTotalNumberOfQueues(){
    int total = 0;
    for (auto & d_key_iter: queues){
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_key_iter.second.hasKey(a_ind)){
                ++total;
            }
        }
    }
    return total;
}

void DepArrQueues::redirectZeroProbabilityTo(Pair d_zero, int capacity){
    double prob;
    int existing;
//     Pair d_zero = Pair(0, 0);
    std::map<int, double> total_zero;
    for (auto d_iter = queues.begin(); d_iter != queues.end(); /* no increment*/){
        if (d_iter->first == d_zero){
            ++d_iter;
            continue;
        }
//         std::cout << "processing " << d_iter->first << "\n";
        existing = 0;
        for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
            if (d_iter->second.hasKey(a_ind)){
                prob = d_iter->second[a_ind][0];
                addValue(total_zero, a_ind, prob);
                d_iter->second[a_ind].setProbability(0, 0);
                if (d_iter->second[a_ind].hasZeroProbability()){
                    d_iter->second.eraseItem(a_ind);
                } else {
                    ++existing;
                }
            }
        }
        if (existing > 0){
            ++d_iter;   
        } else {
            d_iter = queues.erase(d_iter);
        }
    }
//     std::cout << "total zero: " << total_zero << "\n";
    for (auto & a_ind_iter: total_zero){
        int a_ind = a_ind_iter.first;
        if (queues[d_zero].hasKey(a_ind)){
            queues[d_zero][a_ind].addProbability(0, a_ind_iter.second);
        } else {
            queues[d_zero].emplace(a_ind, capacity, a_ind_iter.second);
        }
    }
}


void DepArrQueues::eraseAllQueues(){
    queues.clear();
}



