#ifndef DEPARRQUEUES_H
#define DEPARRQUEUES_H
#include "queue.h"
#include "abpmap.h"
#include <unordered_map>


namespace std{
    template <>
    struct hash<Pair>{
        std::size_t operator() (const Pair & d_key) const{
//             return 1;
            return (d_key.first + 30) * 30 + d_key.second + 2;
        }
    };
}


class ArrQueues{
public:
    ArrQueues() {}
    ABPMap<Queue> queues;
    Queue& operator[](char a_key){ return queues[a_key]; }
    const Queue& operator[](char a_key) const { return queues[a_key]; }
    Queue& operator[](int a_key){ return queues[a_key]; }
    const Queue& operator[](int a_key) const { return queues[a_key]; }
    bool hasAnyKey() const { return queues.hasAnyKey(); }
    bool hasKey(int ind) const {return queues.hasKey(ind);}
    bool hasKey(char a_key) const {return queues.hasKey(a_key);} 
    void addQueueOrInsert(int ind, Queue & queue);
    void addQueueOrInsert(char key, Queue & queue) = delete;
    void eraseItem(int ind) {queues.eraseItem(ind);}
    void eraseItem(char key) = delete;
    void insert(int ind, const Queue & queue);
    void insert(char a_key, const Queue & queue);
    template<typename... Args> bool emplace(int ind, const Args&... args) {return queues.emplace(ind, args...);}
    template<typename... Args> void justEmplace(int ind, const Args&... args) {queues.justEmplace(ind, args...);}
    template<typename... Args> void justEmplace(char key, const Args&... args) = delete;
    template<typename... Args> bool emplace(char key, const Args&... args) = delete;
//     const Queue& at(char a_key) const { return queues.at(a_key);}
};

class DepArrQueues{
public:
    std::map<Pair, ArrQueues> queues;
    ArrQueues& operator[](Pair d_key){ return queues[d_key]; }
//     const ArrQueues& at(Pair d_key) const { return queues.at(d_key);}
    void addKey(Pair d_key) {this->operator[](d_key);}
    void addQueues(DepArrQueues & to_add);
    void removeDownstreamParts(int second, int downstream_second);
    void removeDownstreamParts(int m_second);
    double getTotalProbability() const;
    void renormaliseQueues(double total);
    void printQueues() const;
    void deleteEmptyQueues();
    
    void eraseItem(Pair d_key);
    void eraseAllQueues();

    double getAverageQueue();
    double getAverageSquareQueue(int start_from = 0);
    PairMapChar getProbabilitiesPerAKeys();
    
    int getTotalNumberOfQueues();
    void redirectZeroProbabilityTo(Pair d_key, int capacity);
    template<typename... Args> bool emplace(Pair d_key, int ind, const Args&... args) { 
        if (doesEmplaceMakeSense(args...)){
//             return queues[d_key].emplace(ind, args...);
            auto d_iter = queues.find(d_key);
            if (d_iter == queues.end()){
                queues[d_key].justEmplace(ind, args...);
            } else {
                d_iter->second.justEmplace(ind, args...);
            }
            return true;
        }
        return false;
    }
    template<typename... Args> bool emplace(Pair d_key, char key, const Args&... args) = delete;
};

#endif // DEPARRQUEUES_H
