#include <iostream>
#include <algorithm>
#include <cmath>
#include "departurechain.h"
#include "general.h"
#include "state.h"


double getLastDepartureProbability(int green_time, int & capacity_ceil, int & last_dep_index){
    capacity_ceil = getCapacityCeil(green_time);
//     int i = 0;
    double capacity = getSumoCapacity(green_time);
    last_dep_index = int(capacity);
    double last_departure_probability = capacity - int(capacity);
//     if (last_departure_probability > 1){
//         switchStyle("red");
//         std::cout << green_time << " " << capacity_ceil << " " << sumo_capacity[green_time - 1] << "\n";
//         std::cout << "Warning: incorrect last departure probability!\n";
//         switchStyle("reset");
//     }
    if (last_departure_probability < almost_zero && capacity_ceil > 0) {
//         capacity_ceil -= 1;
        last_dep_index -= 1;
        last_departure_probability = 1;
    }
//     switchStyle("red");
//     std::cout << last_departure_probability << " " << capacity_ceil << " " << last_dep_index << " " << green_time << "\n";
//     switchStyle("reset");
    return last_departure_probability;
}

// bool nextPlusShouldBeRed(int second, int next_red) {
//     // second = last second of a departure state
//     return false;
// //     return second + extra_green + 1 >= next_red;
// }

int DepartureChain::createFixedStates(int second, int number_of_departures, bool is_empty, bool current /*=false*/, int suggested_red /*=-1*/){
    // return next red
    int red = 0;
    Pair a, b;
    for (int k = 0; k < number_of_departures; ++k){
        a = Pair(second, k);
        b = Pair(second, k + 1);
        if (isInMap(chain, a)){
            chain.erase(a); // add new 
        }
        int dep_moment = second + int(getDepartureMoment(k));
        if (k == number_of_departures - 1 && !current){
            red = dep_moment + 1;
            b = Pair(red, -1);                
        }
//         addState(a, dep_moment);
//         addTransition(a, b, 1, 1);
//         default_transitions[a] = b;
        addStateAndTransition(a, dep_moment, b, 1, 1);
        if (!is_empty)
            chain[a].addProbability(1);
    }
    if (current) {
        // add extra state for last departure
        a = Pair(second, number_of_departures);
        red = suggested_red;
        b = Pair(suggested_red, -1);
//         addState(a, suggested_red - 1); 
//         addTransition(a, b, 1, 0);
//         default_transitions[a] = b;
        addStateAndTransition(a, red - 1, b, 1, 0);
        if (!is_empty)
            chain[a].addProbability(1);
    }
    return red;
}

void DepartureChain::createStates(int second, int next_red){
    int capacity_ceil, last_dep_index;
    Pair a, b;
    int rest_green = next_red - second;
    double last_departure_probability = getLastDepartureProbability(rest_green, capacity_ceil, last_dep_index);
// //     std::cout << "create states\n";
//     switchStyle("blue");
//     std::cout << "Creating states for " << second << ", next red is " << next_red << ", capacity_ceil is " << capacity_ceil 
//               << ", last departure probability " << last_departure_probability << ", last_dep_index " << last_dep_index << "\n"; 
//     switchStyle("reset");
    for (int k = 0; k < capacity_ceil; ++k){ // iterate through possible departure numbers
//         std::cout << k << "\n";
        a = Pair(second, k);
        b = Pair(second, k + 1);
//         c = b;
        if ((k == capacity_ceil - 1) && (rest_green == int(getDepartureMoment(capacity_ceil - 1)) + 1))
            b = Pair(next_red, -1);
//         std::cout << "added\n";
        // adding transitions
        double departure_probability = 1.;
        if (k == last_dep_index)
            departure_probability = last_departure_probability;
        else if (k > last_dep_index)
            departure_probability = 0;
//         if (k >= last_dep_index && nextPlusShouldBeRed(second + int(dep_moments[k]), next_red))
//             c = Pair(next_red, -1);
//         else
//             c = b;
//         addState(a, second + int(getDepartureMoment(k)));
//         addTransition(a, b, 1, departure_probability); 
// //         if ((k == capacity_ceil - 1) && rest_green > int(dep_moments[capacity_ceil - 1]) + 1)
// //             b = Pair(second, k + 1);
//         default_transitions[a] = b;
        addStateAndTransition(a, second + int(getDepartureMoment(k)), b, 1, departure_probability);
    }
    // last green state
    if (capacity_ceil == 0 || rest_green > int(getDepartureMoment(capacity_ceil - 1)) + 1){
        a = Pair(second, capacity_ceil);
        // adding transitions
        b = Pair(next_red, -1);
//         addState(a, next_red - 1);
//         addTransition(a, b, 1, 0);
//         default_transitions[a] = b;
        addStateAndTransition(a, next_red - 1, b, 1, 0);
    }
}
/*
Pair DepartureChain::getNextRedState(Pair d_key){
    Pair next_d_key = default_transitions[d_key];
    

}*/


void DepartureChain::addStateAndTransition(Pair d_key, int last, Pair next_d_key, double prob, double dep_prob){
    addState(d_key, last);
    addTransition(d_key, next_d_key, prob, dep_prob);
    default_transitions[d_key] = next_d_key;
}

DepartureChain::DepartureChain(int duration):
        cycle(-2), green_offset(-1), green_time(-1){
    // Make a simple departure chain for a split lane
    Pair a;
    int total_length = duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
    next_green_red.reserve(total_length);
    next_green_red.resize(total_length - 1, -(total_length - 1)); // next red is at second total_length - 1
    next_green_red.push_back(total_length); // next green is at second total_length
    a = Pair(0, 0); // the only state needed
    addState(a, total_length - 2);
}


int DepartureChain::createStates(const std::vector<Pair>& green_periods_so_far, int size, bool is_empty){
    int red = -1;
    for (int i = 0; i < size; ++i){
        Pair pair = green_periods_so_far[i];
        int s = pair.first;
        int k = pair.second;
        if (red >= 0)
            addStateAndTransition(Pair(red, -1), s - 1 - extra_green, Pair(s, 0), 1, 0);
        red = createFixedStates(s, k, is_empty);
    }
    return red;
}



void updateLaneStateIfNeeded(Pair & lane_state, Pair & queue_state, int starting_second, bool is_empty){
    if (lane_state.second >= 0) {
        int last_second = lane_state.first + int(getDepartureMoment(lane_state.second));
        if (last_second < starting_second) {
            if (queue_state.first == lane_state.first)
                queue_state.first += starting_second - last_second;
            lane_state.first += starting_second - last_second;
        }
        if (lane_state.second > 0 && !is_empty) {
            int last_second_previous_departure = lane_state.first + int(getDepartureMoment(lane_state.second - 1));
            if (last_second_previous_departure >= starting_second){
                if (queue_state.first == lane_state.first)
                    queue_state.first -= last_second_previous_departure - starting_second + 1;
                lane_state.first -= last_second_previous_departure - starting_second + 1;
            }
        }
    }
}


void DepartureChain::addProbabilityToStates(Pair queue_state, Pair lane_state){
    Pair state = queue_state;
//     bool add_one = true;
//     printStates();
//     std::cout << queue_state << " to " << lane_state << "\n";
    while (state != lane_state) { // && add_one) {
        chain[state].addProbability(1);
//         std::cout << state << "\n";
//         add_one = (chain[state].getLastSecond() < starting_second);
//         if (!add_one)
//             lane_state = state; // change lane state so that last-arrivals are starting with the right state
        state = default_transitions[state];
        if (state.first > lane_state.first)
            break;
    }
//     if (add_one)
    chain[lane_state].addProbability(1);
}


DepartureChain::DepartureChain(int duration, std::vector<Pair> & green_periods_so_far, int intention, int starting_second, 
                               Pair & lane_state, Pair & queue_state, bool is_empty):
        cycle(-2), green_offset(-1), green_time(-1){
    // Make a simple departure chain for a split lane
    Pair a;
    int total_length = starting_second + duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
    int red = -1, size = int(green_periods_so_far.size());
    
//     std::cout << lane_state << "\n";
//     std::cout << green_periods_so_far << "\n";
    
    next_green_red.reserve(total_length);
    next_green_red.resize(total_length - 1, -(total_length - 1)); // next red is at second total_length - 1
    next_green_red.push_back(total_length); // next green is at second total_length
    a = Pair(0, 0); // the only state needed if there were no blocking lately
    addState(a, total_length - 2);
    
    red = createStates(green_periods_so_far, size - 1, is_empty);
    
    if (lane_state.first > 0 || lane_state.second > 0) {
        updateLaneStateIfNeeded(lane_state, queue_state, starting_second, is_empty);
//     if (size > 0){
        // make many states 
//         pair = green_periods_so_far[size - 1];
        int s = lane_state.first;  // this may be updated
//         if (pair.first != s){
//             std::cout << lane_state << " " << pair << " " << green_periods_so_far << " " << queue_state << "\n";
//             throw "Strange initial conditions for a split lane\n";
//         }
            
//         k = pair.second;
        if (red >= 0)
            addStateAndTransition(Pair(red, -1), s - 1 - extra_green, Pair(s, 0), 1, 0);
        createStates(s, -next_green_red[s]);
    }
    if (!isInMap(chain, lane_state)){
//         createStates(lane_state.first, -next_green_red[lane_state.first]);
        std::cout << lane_state << " " << starting_second << " " << green_periods_so_far << "\n";
        throw "The lane_state should be already in the map, isn't it?\n";
    }
    if (!is_empty)
        addProbabilityToStates(queue_state, lane_state);
//         chain[lane_state].addProbability(1.);
    alternative_routing[lane_state.first][intention] = 1; // TODO: make it possible to use alternative routing for non-zero states && negative intention
    if (lane_state.second < 0) {
        std::cout << lane_state << "\n";
        throw "Incorrect initial lane state. For a split lane it is (x, -1)!\n";
    } else if (lane_state.first == 0 && lane_state.second == 0 && !is_empty){
        std::cout << lane_state << "\n";
        throw "Incorrect initial lane state. For a non-empty split lane it is (0, 0)!\n";
    }
//     printStates();
}



void updateNextGreenRed(std::vector<int> & next_green_red, int & n_gr_index, int start, int end, int total_length){
    while (n_gr_index < start){
        next_green_red[n_gr_index] = start; // positive index for green
        ++n_gr_index;
    }
    while (n_gr_index < end + extra_green && n_gr_index < total_length){
        next_green_red[n_gr_index] = -end - extra_green; // negative for red periods
        ++n_gr_index;
    }        
}


void truncateStartsAndEnds(int total_duration, std::vector<int> & starts, std::vector<int> & ends){
    int size = starts.size();
    int index_to_truncate = size;
    for (int i = size - 1; i >= 0; --i) {
        if (starts[i] >= total_duration)
            index_to_truncate = i;
        else
            break;
    }
    if (index_to_truncate < size){
        starts.resize(index_to_truncate);
        ends.resize(index_to_truncate);
    }
}

DepartureChain::DepartureChain(int duration, std::vector<int> & starts, std::vector<int> & ends):
        cycle(-1), green_offset(-1), green_time(-1){
    // Make a (predefined) departure chain for an arbitrary green phase schedule
//     std::cout << "in constructor\n";
    truncateStartsAndEnds(duration, starts, ends);
    int start, index = 0;
    Pair a;
    next_green_red.resize(duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION);
    if (starts.size() == 0){
        throw "Incorrect vectors of green-period starts and ends.\n";
    }
    start = starts[0];
    if (start > 0){
        a = Pair(0, -1);
//         addState(a, start - 1 - extra_green);  // add last second just before the next green period // - extra_green for extra free departures just before start of the green period
//         addTransition(a, Pair(start, 0), 1, 0);
//         default_transitions[a] = Pair(start, 0);
        addStateAndTransition(a, start - 1 - extra_green, Pair(start, 0), 1, 0);
    }
//     std::cout << "added first state\n";
    for (unsigned int i = 0; i < starts.size(); ++i){
        int end = ends[i];
        updateNextGreenRed(next_green_red, index, start, end, duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION);
        createStates(start, end); // creating green-time states
        // red period state
        a = Pair(end, -1);
        if (i + 1 < starts.size())
            start = starts[i + 1];
        else
            start = duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
//         addState(a, start - 1 - extra_green);  // - extra_green to get first extra_green seconds for free departures
//         addTransition(a, Pair(start, 0), 1, 0);
//         default_transitions[a] = Pair(start, 0);
        addStateAndTransition(a, start - 1 - extra_green, Pair(start, 0), 1, 0);
//         std::cout << "added state " << i << "\n";
    }
    while (index < duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION){
        next_green_red[index] = duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
        ++index;
    }
//     std::cout << "out of constructor\n";
//     std::cout << starts << "\n" << ends << "\n";
//     std::cout << next_green_red << "\n";
//     throw "BU!\n";
}



DepartureChain::DepartureChain(int duration, std::vector<int> & starts, std::vector<int> & ends, 
                               std::vector<Pair>& green_periods_so_far, int intention, int starting_second, Pair & lane_state, Pair & queue_state, bool is_empty):
            cycle(-1), green_offset(-1), green_time(-1){
//     std::cout << "constructing departure chain for a traffic lane\n";
    int size = green_periods_so_far.size();
    int index = 0, start = -1, end = -1, n_gr_index = 0;
    int total_length = starting_second + duration + EXTRA_TIME_IN_THE_END_OF_SIMULATION;
//     std::cout << lane_state << " " << starting_second << "\n" << starts << " " << ends << "\n";
//     std::cout << green_periods_so_far << "\n";
    
    truncateStartsAndEnds(starting_second + duration, starts, ends);
    
    // reserve/ resize next_green_red array
    next_green_red.resize(total_length);
    
    // create states from known green_periods
    int red = createStates(green_periods_so_far, lane_state.second == -1 ? size : size - 1, is_empty);  // if now red use all data else don't use the last one
    // fill first next_green_red
    if (starts.size() == 0){
        throw "Incorrect vectors of green-period starts and ends.\n";
    }
    while (index < int(starts.size())){
        start = starts[index];
        end = ends[index];
        updateNextGreenRed(next_green_red, n_gr_index, start, end, total_length);
        if (start >= lane_state.first) // first period starting after the last one
            break;
        ++index;
    }
//     std::cout << starts << "\n" << start << "\n";
//     std::cout << "red: " << red << "\n";
    if (index >= int(starts.size())){ // there are no new green phases (this is strange, and not good but not very bad)
        start = total_length;
//         throw "Not enough green periods data!\n";
    }
    if (red == -1 && start > 0){ // no green period so far => add red period until start
        addStateAndTransition(Pair(0, -1), start - 1 - extra_green, Pair(start, 0), 1, 0);
    } else if (red >= 0 && green_periods_so_far[0].first > 0) {
        addStateAndTransition(Pair(0, -1), green_periods_so_far[0].first - 1 - extra_green, Pair(green_periods_so_far[0].first, 0), 1, 0);
    }
//     std::cout << "added before lane state\n";
    if (lane_state.second >= 0) {  // now still green => create either fixed states or more states with normal departure probs
        // add current green period
        int s = start;
        int k = lane_state.second;
        if (size > 0) {
            s = green_periods_so_far[size - 1].first;  // size should be here more than 0
            k = green_periods_so_far[size - 1].second;
        }
        if (s != lane_state.first){
            std::cout << lane_state << " " << s << " " << green_periods_so_far << " " << starts << " " << ends << "\n";
            throw "Strange green periods data. Incorrect green beginning.\n";
        }
        
        updateLaneStateIfNeeded(lane_state, queue_state, starting_second, is_empty);
        s = lane_state.first;
        start = s;
        if (red >= 0)
            addStateAndTransition(Pair(red, -1), s - 1 - extra_green, Pair(s, 0), 1, 0);
        double capacity = getSumoCapacity(end - start);
        if (k <= capacity + small_eps) {  /// k <= capacity
            createStates(s, end);
        } else { //more departures than on average
            createFixedStates(s, k, true, end);
        }
        updateNextGreenRed(next_green_red, n_gr_index, start, end, total_length);
        red = end;
        ++index;
        if (index < int(starts.size())) {
            start = starts[index];
            addStateAndTransition(Pair(red, -1), start - 1 - extra_green, Pair(start, 0), 1, 0);
        } else {
            start = total_length;
            addState(Pair(red, -1), total_length - 1);
        }
    } else {
        // add this state
        if (!isInMap(chain, lane_state))
            addStateAndTransition(lane_state, start - 1 - extra_green, Pair(start, 0), 1, 0);
        if (red >= 0 && !isInMap(chain, Pair(red, -1)))
            addStateAndTransition(Pair(red, -1), start - 1 - extra_green, Pair(start, 0), 1, 0);
    }
//     std::cout << "added lane_state\n";
    //add all other green periods
    for (unsigned int i = index; i < starts.size(); ++i){
        end = ends[i];
        updateNextGreenRed(next_green_red, n_gr_index, start, end, total_length);
        createStates(start, end); // creating green-time states
        // red period state
        Pair a(end, -1);
        if (i + 1 < starts.size())
            start = starts[i + 1];
        else
            start = total_length;
        addStateAndTransition(a, start - 1 - extra_green, Pair(start, 0), 1, 0);
    }
    updateNextGreenRed(next_green_red, n_gr_index, total_length, total_length, total_length); // the rest of the time is red
//     std::cout << "almost finished constructing departure chain for a traffic lane\n";
    if (!is_empty)
        addProbabilityToStates(queue_state, lane_state);
//         chain[lane_state].addProbability(1.);
    alternative_routing[lane_state.first][intention] = 1; // TODO: make it possible to use alternative routing for non-zero states && negative intention
//     std::cout << "finished constructing departure chain for a traffic lane\n";
//     std::cout << "lane_state " << lane_state << "\n";
//     std::cout << queue_state << ", " << green_periods_so_far << "\n";
//     printStates();
}


DepartureChain::DepartureChain(int duration, int cycle, int offset, int red, int green): 
        cycle(cycle), green_offset((offset + red) % cycle), green_time(green + extra_green){
    // Make a predefined departure chain for a fixed cycle 
    int green_beginning = 0, next_red = green_offset + green_time;
//     int capacity_ceil = 0;
//     double last_departure_probability = 0., departure_probability;
    
    if (!(((green_offset + green_time > cycle) && (cycle >= green_offset)) || green_offset == 0)){
        Pair a(0, -1);
        green_beginning = green_offset;
//         addState(a, green_offset - 1 - extra_green);  // add last second just before the next green period // - extra_green for extra free departures just before start of the green period
//         addTransition(a, Pair(green_beginning, 0), 1, 0);
//         default_transitions[a] = Pair(green_beginning, 0);
        addStateAndTransition(a, green_beginning - 1 - extra_green, Pair(green_beginning, 0), 1, 0);
    } else if (green_offset != 0){
        green_beginning = green_offset - cycle;  // < 0
    }
    // find capacity and probability of the last departure
//     last_departure_probability = getLastDepartureProbability(green_time, capacity_ceil);
    // iterate through all green times before duration stops
    while (green_beginning < duration + 20){
        createStates(green_beginning, next_red);
//         for (int k = 0; k < capacity_ceil; ++k){ // iterate through possible departure numbers
//             a = Pair(green_beginning, k);
//             b = Pair(green_beginning, k + 1);
//             if ((k == capacity_ceil - 1) && (green_time == int(dep_moments[capacity_ceil - 1]) + 1))
//                 b = Pair(next_red, -1);
//             addState(a, green_beginning + int(dep_moments[k]));
//             // adding transitions
//             departure_probability = 1.;
//             if (k == capacity_ceil - 1)
//                 departure_probability = last_departure_probability;
//             addTransition(a, b, 1, departure_probability);
//             default_transitions[a] = b;
//         }
//         // last green state
//         if (green_time > int(dep_moments[capacity_ceil - 1]) + 1){
//             a = Pair(green_beginning, capacity_ceil);
//             addState(a, next_red - 1);
//             // adding transitions
//             b = Pair(next_red, -1);
//             addTransition(a, b, 1, 0);
//             default_transitions[a] = b;
//         }
        // red period state
        Pair a(next_red, -1);
        green_beginning = next_red + cycle - green_time;
//         addState(a, green_beginning - 1 - extra_green);  // - extra_green to get first extra_green seconds for free departures
//         addTransition(a, Pair(green_beginning, 0), 1, 0);
//         default_transitions[a] = Pair(green_beginning, 0);
        addStateAndTransition(a, green_beginning - 1 - extra_green, Pair(green_beginning, 0), 1, 0);
        next_red += cycle;
    }
//     printStates();
}



void DepartureChain::addState(int t, int k){
    addState(Pair(t, k));
}


void DepartureChain::addState(Pair pair){
    chain.emplace(pair, true);
}

void DepartureChain::addState(Pair pair, int last){
    addState(pair);
//     std::cout << "addState\n";
    chain[pair].updateLastSecond(last);
//     std::cout << "updated last second\n";
}


// void DepartureChain::addState(Pair pair, int last, bool is_green_begin){
//     chain[pair] = DepartureState(is_green_begin);
//     chain[pair].updateLastSecond(last);
// }

void DepartureChain::addTransition(int t1, int k1, int t2, int k2, double probability, double departure_probability){
    addTransition(Pair(t1, k1), Pair(t2, k2), probability, departure_probability);
}


void DepartureChain::addTransition(Pair pair1, Pair pair2, double probability, double departure_probability){
    chain[pair1].addTransition(pair2, probability, departure_probability);
}

void DepartureChain::redirectDefaultTransition(Pair pair1, Pair pair2, double probability){
    // here, default transitions plus
//     std::cout << "redirect default transitions of " << pair1 << "\n";
    chain[pair1].transitions.redirectTransition(getDefaultPlusTransition(pair1), pair2, probability);
//     switchStyle("yellow");
//     std::cout << pair1 << " " << default_transitions[pair1] << " -> " << pair2 << " " << probability <<"\n";
//     std::cout << "redirected\n";
//     chain[pair1].transitions.printTransitions();
//     std::cout << chain[pair1].getProbability() << "\n";
//     switchStyle("reset");
// the probabilities of the states are altered in a different place
//     chain[pair2].addProbability(probability); // do not remove from default of pair 1 because it is computed later
}


void DepartureChain::printStates(){
    std::cout << "States of the departure chain\n";
    for (auto & d_iter : chain){
        std::cout << "\t" << d_iter.first << ": last second " << d_iter.second.getLastSecond() << ", transition probabilities:\n";
        d_iter.second.transitions.printTransitions();
        std::cout << "departure probabilities during transitions:\n";
        d_iter.second.departure_probabilities.printTransitions();
        std::cout << "probability: " << d_iter.second.getProbability() << "\n";
    }
}


Pair DepartureChain::getFirstState(){
    return chain.begin()->first;
}

bool DepartureChain::isFreeDepartureAllowed(Pair d_key, int second){
    return (d_key.first <= second && d_key.second > 0) || (isGreenBegin(d_key.first) && d_key.second == 0 && d_key.first <= second + extra_green);
}

std::map<Pair, double> DepartureChain::getTransitions(int t, int k){
    return getTransitions(Pair(t, k));
}

std::map<Pair, double> DepartureChain::getTransitions(Pair pair){
    return chain[pair].transitions.getTransitions();
}

double DepartureChain::getDefaultTransitionProbability(Pair d_key){
    return chain[d_key].transitions.getTransition(default_transitions[d_key]);
}


std::map<Pair, double> DepartureChain::getDepartureProbabilities(Pair pair){
    return chain[pair].departure_probabilities.getTransitions();
}

double DepartureChain::getDepartureProbability(Pair pair){
    // this is departure probability for the default transition
    if (isGreen(pair.first)){ // this state is green
        int capacity_ceil = 0, last_dep_index;
        int new_green_time, second = pair.first;
        if (cycle > 0) {
            new_green_time = green_time - getGreenSecond(second);
        } else {
            new_green_time = -getNextGreenRed(second) - second;
        }
        double last_departure_probability = getLastDepartureProbability(new_green_time, capacity_ceil, last_dep_index);
        int k = pair.second;
        if (k == last_dep_index)
            return last_departure_probability;
        else if (k < last_dep_index)
            return 1;
    }
    return 0;
}

Pair DepartureChain::getDefaultZeroTransition (Pair d_key){
    return default_transitions[d_key]; // normal default transition
}

Pair DepartureChain::getDefaultPlusTransition (Pair d_key){
    return default_transitions[d_key]; // the different definition is not used anymore
//     Pair next_d_key = default_transitions[d_key]; //, other_d_key;
// //     if (getDepartureProbability(next_d_key) < almost_zero && next_d_key.second >= 0){ // maybe a red state is needed
// //         other_d_key = next_d_key;
// //         int last_second = chain[d_key].getLastSecond();
// //         while (other_d_key.second >= 0 && nextPlusShouldBeRed(last_second, other_d_key.first)) // next state is green but without departures, than skip it
// //             other_d_key = default_transitions[other_d_key]; // this will be next red state
// //         if (other_d_key.second == -1 && nextPlusShouldBeRed(last_second, other_d_key.first)){ // is red
// //             // should change if nextPlusShouldBeRed is false
// //             next_d_key = other_d_key;
// //         }
// //     }
//     return next_d_key;
}

Pair DepartureChain::getDefaultZeroTransition(int second, Pair d_key){
    if (chain[d_key].getLastSecond() <= second)
        return getDefaultZeroTransition(d_key);
    else 
        return d_key;
}

Pair DepartureChain::getDefaultPlusTransition(int second, Pair d_key){
    if (chain[d_key].getLastSecond() <= second)
        return getDefaultPlusTransition(d_key);
    else 
        return d_key;
}


void DepartureChain::updateRoutingFlow(Pair d_key, int lane_id, double value){
    // Add value to routing probabilities
    // and to the probability of state d_key
//     std::cout << d_key << " " << value << "\n";
    if (!isInMap(chain, d_key)){
        createUnblockingState(d_key.first);
    }
    chain[d_key].addProbability(value);
    int second = d_key.first;
    addValue(alternative_routing, second, lane_id, value);
}


int DepartureChain::getNextGreenRed(int second){
    // get next_green_red[second] if such a second exists in the vector,
    // otherwise, "switch" to cyclic structure with 1 second of green and 1 second of red
    // starting with green second
    if (second < int(next_green_red.size()))
        return next_green_red[second];
    if (second < 0)
        throw "Negative second in getNextGreenRed\n";
    // green second for 0 mod 2 and red for 1 mod 2
    return ((second - next_green_red.size()) % 2 == 0) ? (- second - 1): second + 1;
}


void DepartureChain::createUnblockingState(int second){
//     int capacity_ceil = 0, last_dep_index;
    int next_red; 
//     switchStyle("red");
//     std::cout << "Creating unblocking state at second " << second << "\n";
//     switchStyle("reset");
    if (cycle > 0){
        if (getGreenSecond(second) < green_time)
            next_red = second - getGreenSecond(second) + green_time;
        else
            next_red = second - getGreenSecond(second) + green_time + cycle;
    } else {
        next_red = getNextGreenRed(second); // this is next period red or green
        if (next_red > 0) // actually next is green
            next_red = getNextGreenRed(next_red); // this should be red 
        next_red = - next_red;
    }
//     int rest_green = next_red - second;
    createStates(second, next_red);
//     Pair a, b;
//     double last_departure_probability = getLastDepartureProbability(rest_green, capacity_ceil, last_dep_index);
//     double departure_probability;
//     
//     std::cout << capacity_ceil << " " << next_red << " " << rest_green << " "<< last_departure_probability << "\n"; 
//     for (int k = 0; k < capacity_ceil; ++k){ // iterate through possible departure numbers
//         a = Pair(second, k);
//         b = Pair(second, k + 1);
//         if ((k == capacity_ceil - 1) && (rest_green == int(dep_moments[capacity_ceil - 1]) + 1))
//             b = Pair(next_red, -1);
//         addState(a, second + int(dep_moments[k]));
//         // adding transitions
//         departure_probability = 1.;
//         if (k == capacity_ceil - 1)
//             departure_probability = last_departure_probability;
//         addTransition(a, b, 1, departure_probability);
//         default_transitions[a] = b;
//     }
//     // last green state
//     if (capacity_ceil == 0 || rest_green > int(dep_moments[capacity_ceil - 1]) + 1){
//         a = Pair(second, capacity_ceil);
//         addState(a, next_red - 1);
//         // adding transitions
//         b = Pair(next_red, -1);
//         addTransition(a, b, 1, 0);
//         default_transitions[a] = b;
//     }
}



void DepartureChain::updateStates(const std::map<Pair, std::map<Pair, double> >& tr_flows_plus, 
                                  const std::map<Pair, std::map<int, double> >& unblocking_d_keys_probabilities){
//     std::cout << "updateStates\n";
//     std::cout << unblocking_d_keys_probabilities << "\n";
//     std::cout << tr_flows_plus << " <- tr_flows_plus\n";
    for (auto & d_iter: unblocking_d_keys_probabilities){
        for (const auto & iter: d_iter.second){
//             switchStyle("blue");
//             std::cout << "updating routing flow\n";
//             std::cout << d_iter.first << " " << iter.first << " " << iter.second <<"\n";
//             switchStyle("reset");
            updateRoutingFlow(d_iter.first, iter.first, iter.second);
        }
    }
    
    //tr_flows_plus update transition probabilities for states (for positive queue length)
//     double state_plus_prob, default_transition_prob;
    for (const auto & d_iter_from: tr_flows_plus){
//         std::cout << d_iter_from.first << "\n";
        double state_plus_prob = chain[d_iter_from.first].getProbability();
//         std::cout << state_plus_prob << "\n";
        for (auto & d_iter_to: d_iter_from.second){
            double default_transition_prob = getDefaultTransitionProbability(d_iter_from.first); // it may change if there are multiple redirections
//             std::cout << default_transition_prob << " <- default transition\n";
//             std::cout << d_iter_from.first << " " << d_iter_to.first << " " << d_iter_to.second << " " << state_plus_prob << "\n";
            if ((d_iter_to.second > state_plus_prob * default_transition_prob + small_eps)){ // && (d_iter_to.second > eps)){
//                 printStates(); 
                std::cout << d_iter_from.first << " " << d_iter_to.first << " " << d_iter_to.second << " " << state_plus_prob << "\n";
                std::cout << tr_flows_plus << "\n";
                std::cout << unblocking_d_keys_probabilities << "\n";
                throw "updateStates of a departure chain: wrong transition flow!\n";
            }
            if (state_plus_prob > almost_zero){
                redirectDefaultTransition(d_iter_from.first, d_iter_to.first, 
                                          std::min(d_iter_to.second, default_transition_prob * state_plus_prob) / state_plus_prob);
            }
        }
    }
//     std::cout << "states updated\n";
}


std::map<int, double> DepartureChain::getRoutingProbabilities(Pair d_key, const std::map<int, double>& routing_probabilities){
    std::map<int, double> result = alternative_routing[d_key.first];
    if (d_key.second != 0)
        return routing_probabilities; 
    if (isGreenBegin(d_key.first)){ // green begin state
        double total = sumMapValues(result);
        if (total < 1){
            for (const auto & iter: routing_probabilities){
                addValue(result, iter.first, iter.second * (1 - total));
            }
        }
    } else { // is unblocking state
        double prob = chain[d_key].getProbability();  // should be positive since otherwise the state won't exist
        if (prob < almost_zero) {
            if (sumMapValues(result) > almost_zero){
                std::cout << prob << "\n";
                std::cout << d_key << "\n";
                std::cout << result << "\n";
                throw "Incorrect probability of the departure state!\n";
            } else {
                // doesn't matter as it is very small probability 
            }
        } else {
            for (const auto & iter: result){
                result[iter.first] /= prob;
            }
        }
    }
    return result;
}

double DepartureChain::getRoutingProbabilities(Pair d_key, const std::map<int, double>& routing_probabilities, int u){
    if (d_key.second != 0)
        return routing_probabilities.at(u); 
    int second = d_key.first;
    double result = 0;
    if (isGreenBegin(d_key.first)){ // green begin state
        if (!isInMap(alternative_routing, second))
            return routing_probabilities.at(u);
        std::map<int, double> routing = alternative_routing[d_key.first];
        result = alternative_routing[d_key.first][u];
        double total = sumMapValues(routing);
        return result + routing_probabilities.at(u) * (1 - total);
    } else { // is unblocking state
        if (!isInMap(alternative_routing, second))
            return 0;
        double prob = chain[d_key].getProbability();  // should be positive since otherwise the state won't exist
        double gamma_prob = alternative_routing[second][u];
        if (prob < almost_zero) {
            if (gamma_prob > almost_zero){
                std::cout << prob << "\n";
                std::cout << d_key << "\n";
                std::cout << result << "\n";
                throw "Incorrect probability of the departure state!\n";
            } else {
                // doesn't matter as it is very small probability 
                return 0;
            }
        } else {
            return alternative_routing[second][u] / prob;
        }
    }
}


bool DepartureChain::isGreenBegin(int second){
    if (cycle > 0){
        return getGreenSecond(second) == 0;
    } else {
        Pair first_state = getFirstState();
        return (second > 0 && getNextGreenRed(second - 1) == second) || (second == 0 && first_state.first == 0 && first_state.second == 0);
    }
}


bool DepartureChain::isGreen(int second){
    if (cycle > 0)
        return getGreenSecond(second) < green_time;
    else 
        return getNextGreenRed(second) < 0;
}


bool DepartureChain::isExtendedGreen(int second){
    if (cycle > 0)
        return isGreen(second) || getGreenSecond(second + extra_green) < extra_green;
    else
        return isGreen(second) || (!isGreen(second) && getNextGreenRed(second) <= second + extra_green);
}


int DepartureChain::getExtraSecondsToGreen(int second){
    if (isGreen(second))
        return 0;
    else if (cycle > 0)
        return -getGreenSecond(second + extra_green) + extra_green;
    else
        return getNextGreenRed(second) - second;
}


/*
bool DepartureChain::isDepartureSecond(int second){
    int mod_sec = getGreenSecond(second);
    return mod_sec < green_time && std::find(dep_moments.begin(), dep_moments.end(), mod_sec) != dep_moments.end();
}*/

bool DepartureChain::isDepartureSecond(int second, Pair d_key){
    // check if s_{d}+d_{k_d + 1} == second and it is green
    return isGreen(second) && d_key.second >= 0 && (d_key.first + int(getDepartureMoment(d_key.second)) == second);
}

bool DepartureChain::wasDepartureSecond(int second, Pair d_key){
    // check if s_{d}+d_{k_d + 1} == second and it is green
    return isGreen(second) && d_key.second > 0 && (d_key.first + int(getDepartureMoment(d_key.second - 1)) == second);
}


int DepartureChain::getGreenSecond(int second){
    if (cycle > 0)
        return ((second - green_offset) % cycle + cycle) % cycle;
    else
        throw "Incorrect call, no green second data for arbitrary green phase schedulen\n";
}
/*
int DepartureChain::getDepartureNumber(int second){
    if (isDepartureSecond(second))
        return dep_indexes.at(getGreenSecond(second));
    return -1;
}*/


int DepartureChain::getDepartureNumber(int last_second, Pair d_key){
    if (isGreen(last_second) && d_key.second > 0){ // = 0 is not relevant as it means there were no departures yet
        int second_d = d_key.first;
        int dep_ind = d_key.second - 1;
        if (last_second == second_d + int(getDepartureMoment(dep_ind)))
            return dep_ind;
//         if (isInMap(dep_indexes, last_second - second_d))
//             return dep_indexes.at(last_second - second_d);
    }
    return -1;
}

// int DepartureChain::getNextUnblock(double second){
//     // find next second such that it is a departure
//     if (!isGreen(int(second))){
//         
//     }
// }

int DepartureChain::getNextUnblock(Pair d_key){
    int second = d_key.first, k = d_key.second;
//     std::cout << "getNextUnblock: " << d_key << "\n";
    if ((k < 0) || !isGreen(second + int(getDepartureMoment(k)))) { // red or almost red => no departure
//         std::cout << "(almost) red\n";
        if (cycle > 0) {
//             std::cout << "positive cycle\n";
            return (second - getGreenSecond(second)) + cycle + int(getDepartureMoment(0));
        } else {
//             std::cout << "negative cycle\n";
            if (k < 0){
//                 std::cout << "k < 0: " << next_green_red[second] + int(dep_moments[0]) << "\n";
                return getNextGreenRed(second) + int(getDepartureMoment(0));
            } else { // almost red
//                 std::cout << "almost red: " << -next_green_red[second] << " " << next_green_red[-next_green_red[second]] + int(dep_moments[0]) << "\n";
                return getNextGreenRed(-getNextGreenRed(second)) + int(getDepartureMoment(0));
            }
        }
    } else {
//         std::cout << "green " << second + int(dep_moments[k]) << "\n";
        return second + int(getDepartureMoment(k));
    }
}

int DepartureChain::getFirstNextGreenSecond(int second){
//     std::cout << "getFirstNextGreenSecond " << second << "\n";
    if (isGreen(second)){
//         std::cout << "isGreen\n";
        return second;
    } else if (cycle > 0) {
//         std::cout << "cycle > 0\n";
        return (second - getGreenSecond(second)) + cycle;
    } else {
//         std::cout << "less than size " << next_green_red[second] << "\n";
        return getNextGreenRed(second);
    }
}

inline double getZeroProbability(Pair d_key, std::map<KeyDA, double>& zero_probs){
    double zero_prob = 0;
    for (auto & a_key: {'p', 'a', 'b'}){
        std::map<KeyDA, double>::iterator it = zero_probs.find(KeyDA(d_key, a_key));
        if (it != zero_probs.end()){
            zero_prob += it->second;
        }
    }
    return zero_prob;
}

void DepartureChain::setFirstProbability(int step, std::map<KeyDA, double>& zero_probs, double minus_first_probability){
    if (isGreenBegin(step + extra_green)){
        Pair d_key = Pair(step + extra_green, 0);
        double zero_prob = getZeroProbability(d_key, zero_probs);
        chain[d_key].addProbability(1 - zero_prob - chain[d_key].getProbability() - minus_first_probability); // minus what is already known (due to blocking)
    } else if ((step == 0) && isGreen(step)){
        Pair d_key = getFirstState();
        double zero_prob = getZeroProbability(d_key, zero_probs);
        chain[d_key].addProbability(1 - zero_prob);        
    }
}

Pair getLastDKey(Pair d_key){
    if (d_key.second > 0)
        return Pair(d_key.first, d_key.second - 1);
    // else (d_key.second <= 0)
    std::cout << d_key << "\n";
    throw("No last d_key!\n");
}


void DepartureChain::updateStateProbabilityAfterDepartureSecond(int step){
    Pair new_d_key, d_key;
    for (auto & d_iter: chain){
        d_key = d_iter.first;
        if (d_iter.second.getLastSecond() == step && d_key.second >= 0){ // only green states that just have last second 
            new_d_key = getDefaultPlusTransition(d_key);
//             std::cout << d_key << " to " << new_d_key << "\n";
            chain[new_d_key].addProbability(chain[d_key].getProbability() * chain[d_key].transitions.getTransition(new_d_key));
//             std::cout << "processed transition " << d_key << " to " << new_d_key << "\n";
//             std::cout << "added probability: " << chain[d_key].getProbability() * chain[d_key].transitions.getTransition(new_d_key) << "\n";
        }
    }
}



void DepartureChain::updateStateProbabilityAfterDepartureSecond(std::vector<Pair>& d_key_to_update){
    Pair last_d_key;
    for (auto & d_key: d_key_to_update){
        if (d_key.second >= 0){ // should not be equal to 0 (this will raise a error)
            last_d_key = getLastDKey(d_key);
            chain[d_key].addProbability(chain[last_d_key].getProbability() * chain[last_d_key].transitions.getTransition(d_key));
//             std::cout << "processed transition " << last_d_key << " to " << d_key << " with pr. " << chain[last_d_key].transitions.getTransition(d_key) << "\n";
//             std::cout << "added probability: " << chain[last_d_key].getProbability() * chain[last_d_key].transitions.getTransition( d_key) << "\n";
        }
    }
}


void DepartureChain::updateStateProbabilityFromEmptied(int step, const std::map<KeyDA, double>& just_emptied_probs){
    // iterate through just_emptied_probs to find proper d_keys to look at
    Pair d_key, last_d_key;
    std::map<Pair, double> just_emptied;
    std::map<Pair, double> should_got_emptied;
    double prob;
//     std::cout << just_emptied_probs << " <- updateStateProbability, just emptied probs\n";
    for (const auto & da_iter: just_emptied_probs){
        d_key = da_iter.first.first;
        addValue(just_emptied, d_key, da_iter.second);
    }
//     std::cout << just_emptied << "\n";
    for (auto & d_iter: just_emptied){
        d_key = d_iter.first;
//         std::cout << d_key << "\n";
        if (d_key.second > 0){
            last_d_key = getLastDKey(d_key); // this is a true pair of d_keys: one that just had a departure second and its default transition
//             std::cout << d_key << " was preceeded by " << last_d_key << "\n";
//             std::cout << "with last second " << chain[last_d_key].getLastSecond() << " and step " << step << "\n"; 
//             if (chain[last_d_key].getLastSecond() <= step - 1 && chain[d_key].getProbability() < almost_zero) 
//                 // second condition resembles that of not updating the probability yet (but actually not the same!)
//                 chain[d_key].addProbability(chain[last_d_key].getProbability() * chain[last_d_key].transitions.getTransition(d_key) - d_iter.second);
            if (chain[last_d_key].getLastSecond() <= step - 1){
                if (!chain[d_key].removeProbability(default_transitions[d_key], d_iter.second)){
                    should_got_emptied.clear();
//                     double prob;
                    chain[d_key].printState(d_key);
                    last_d_key = d_key;
//                     chain[last_d_key].printState(last_d_key);
                    while (last_d_key.second > 0) {
                        prob = - chain[last_d_key].getProbability();
                        last_d_key = getLastDKey(last_d_key);
                        chain[last_d_key].printState(last_d_key);
                        prob += chain[last_d_key].getProbability() * getDefaultTransitionProbability(last_d_key);
                        should_got_emptied[last_d_key] = prob;
                    }
                    for (auto & new_d_iter: chain){
                        if (new_d_iter.second.transitions.getTransition(Pair(d_key.first, 0)) > 0){
                            new_d_iter.second.printState(new_d_iter.first);
                        }
                    }
                    std::cout << "should got emptied: " << should_got_emptied << "\n";
                    std::cout << d_key << " <- d_key\n"; // " with probability " << prob << "\n";
                    std::cout << just_emptied << "\nstep is " << step << "\n";
                    printStates();
                    throw "Negative probability after the update\n";
                }
            }
//             std::cout << d_key << " from " << last_d_key << " that has porbability " << chain[last_d_key].getProbability() 
//                       << " and default transition probability " << chain[last_d_key].transitions.getTransition(d_key) << " with " << d_iter.second << " just emptied\n";
    //         std::cout << chain[d_key].getProbability() << "\n";
        }
    }    
//     std::cout << just_emptied << " <- updateStateProbability\n";
}

void DepartureChain::removeStates(std::vector<Pair>& d_keys){
    std::map<int, int> from_which_to_delete;
    // check which also to remove
    int second, k;
    for (Pair d_key: d_keys){
        // add all of the following (if it still exists)
        second = d_key.first;
        k = d_key.second;
        if (isInMap(from_which_to_delete, second)){
            from_which_to_delete[second] = std::min(from_which_to_delete[second], k);
        } else {
            from_which_to_delete[second] = k;
        }
    }
    for (auto pair_iter: from_which_to_delete){
        // remove transition to the first state
        second = pair_iter.first;
        k = pair_iter.second;
        if (k > 0) {
            chain[Pair(second, k - 1)].transitions.removeTransition(Pair(second, k));
            chain[Pair(second, k - 1)].departure_probabilities.removeTransition(Pair(second, k));
        }
        // remove all following states
        int i = 0;
        while (true){
            Pair d_key = Pair(second, k + i);
            if (isInMap(chain, d_key)){
                if (i != 0)
                    d_keys.push_back(d_key); // add so that in other structures these states can be removed
                chain.erase(d_key);
                ++i;
            } else {
                break;
            }
        }
    }
}


