#ifndef DEPARTURECHAIN_H
#define DEPARTURECHAIN_H
#include <vector>
#include <map>
#include "state.h"

//using Pair = std::pair<int, int>;

class DepartureChain{
    int cycle, green_offset, green_time;
    std::map<Pair, DepartureState> chain;  // pair is usually denoted as (t, k)
    std::map<Pair, Pair> default_transitions;
    std::map<int, std::map<int, double>> alternative_routing;
    std::vector<int> next_green_red;

    void addTransition(int t1, int k1, int t2, int k2, double probability, double departure_probability);
    void addTransition(Pair pair1, Pair pair2, double probability, double departure_probability);
    
    void createStates(int second, int next_red);
    int createFixedStates(int second, int number_of_departures, bool is_empty, bool current=false, int suggested_red=-1);
    int createStates(const std::vector<Pair> & green_periods_so_far, int size, bool is_empty);
    void addProbabilityToStates(Pair queue_state, Pair lane_state);
public:
    DepartureChain(int duration, int cycle, int offset, int red, int green);
    explicit DepartureChain(int duration); // departure chain for a split lane
    DepartureChain(int duration, std::vector<Pair> & green_periods_so_far, int intention, int starting_second,
                   Pair & lane_state, Pair & queue_state, bool is_empty);
    DepartureChain(int duration, std::vector<int> & starts, std::vector<int> & ends);
    DepartureChain(int duration, std::vector<int> & starts, std::vector<int> & ends, std::vector<Pair> & green_periods_so_far, 
                   int intention, int starting_second, Pair & lane_state, Pair & queue_state, bool is_empty);
    int getGreenSecond(int second);
    void printStates ();
    void addState(int t, int k);
    void addState(Pair pair);
    void addState(Pair pair, int last);
    void addStateAndTransition(Pair d_key, int last, Pair next_d_key, double prob, double dep_prob);
    bool hasState(Pair pair) { return isInMap(chain, pair); }
//     void addState(Pair pair, int last, bool is_green_begin);
    void redirectDefaultTransition(Pair pair1, Pair pair2, double probability);
    void createUnblockingState(int second);
    
    std::map<Pair, double> getTransitions(int t, int k);
    std::map<Pair, double> getTransitions(Pair pair);
    std::map<Pair, double> getDepartureProbabilities(Pair pair);
    double getDepartureProbability(Pair pair);
    double getStateProbability(Pair pair) { return chain[pair].getProbability();}
    Pair getFirstState();
    bool isFreeDepartureAllowed(Pair d_key, int second);
    int getLastSecond(Pair d_key) {return chain[d_key].getLastSecond(); }
    Pair getDefaultZeroTransition(Pair d_key);
    Pair getDefaultZeroTransition(int second, Pair d_key);
    Pair getDefaultPlusTransition(Pair d_key);
    Pair getDefaultPlusTransition(int second, Pair d_key);
    double getDefaultTransitionProbability(Pair d_key);
    
    std::map<int, double> getRoutingProbabilities(Pair d_key, const std::map<int, double> & routing_probabilities);
    double getRoutingProbabilities(Pair d_key, const std::map<int, double> & routing_probabilities, int u);
    void updateRoutingFlow(Pair d_key, int lane_id, double value);  
    
    void updateStates(const std::map<Pair, std::map<Pair, double>> & tr_flows_plus, const std::map<Pair, std::map<int, double>> & unblocking_d_keys_probabilities);
    
    bool isGreenBegin(int second);
    bool isGreen(int second);
    bool isExtendedGreen(int second);
    bool isDepartureSecond(int second);
    bool isDepartureSecond(int second, Pair d_key);
    bool wasDepartureSecond(int second, Pair d_key);
//     int getDepartureNumber(int second);
    int getDepartureNumber(int second, Pair d_key);
    
    int getFirstNextGreenSecond(int second);
    
//     int getNextUnblock(double second);
    int getNextUnblock(Pair d_key);
    
    int getExtraSecondsToGreen(int second);
    
    void setFirstProbability(int step, std::map<KeyDA, double>& zero_probs, double minus_first_probability);
    void updateStateProbabilityFromEmptied(int step, const std::map<KeyDA, double>& just_emptied_probs);
    void updateStateProbabilityAfterDepartureSecond(std::vector<Pair>& d_key_to_update);
    void updateStateProbabilityAfterDepartureSecond(int step);
    int getNextGreenRed(int second);
    
    bool isSplitLane() {return (cycle == -2);}
    void removeStates(std::vector<Pair>& d_keys);
    //     Pair getNextRedState(Pair d_key);
};

#endif // DEPARTURECHAIN_H
