#include <iostream>
#include <fstream>
#include <chrono>
#include <ctime>
#include <set>
#include "general.h"
#include "state.h"
#include "tail.h"

bool almostEqual(double x, double y, double admisible_error){
    return ((x + admisible_error >= y) && (x <= y + admisible_error));
}

void printWarning(const std::string & text){
    std::cout << "\033[1;31mWARNING: " << text << "\033[0m\n";
}

void printNow(){
    auto now = std::chrono::system_clock::now();
    std::time_t now_time = std::chrono::system_clock::to_time_t(now);
    std::cout << std::ctime(&now_time) << "\n";
}

std::ostream& operator<<(std::ostream& os, const std::set<Pair> & pairs){
    for (auto const& pair: pairs)
        os << pair << " ";
    return os;
}

std::ostream& operator<<(std::ostream& os, const Tail & tail){
    os << "(" << tail.second_from << ", " << tail.a_key_from << ", " << tail.second_to << ", " << tail.a_key_to << ", " << tail.probability << ")";
    return os;
}

// std::ofstream& operator<<(std::ofstream& os, const std::vector<double> & vector){
//     os << "[";
//     for (unsigned int i = 0; i < vector.size(); ++i){
//         os << vector[i];
//         if (i < vector.size() - 1)
//             os << ", ";
//     }
//     os << "]\n";
//     return os;
// }

// template<char, double> std::ostream& operator<< <char, double>(std::ostream&, std::map<char, double, std::less<char>, std::allocator<std::pair<char const, double> > > const&);


int getCapacityCeil(int green_time){
    int capacity_ceil = 0;
    int departure_momements_size = dep_moments.size();
    if (green_time < dep_moments[departure_momements_size - 1]){
        for (int i = 0; i < departure_momements_size; ++i){
            if (dep_moments[i] < green_time)
                ++capacity_ceil;
            else 
                break;
        }
    } else {
        // dep_moments later are approximated as dep_moments[-1] + 2
        int dif = green_time - dep_moments[departure_momements_size - 1];
        capacity_ceil = departure_momements_size + dif / 2;
    }
    return capacity_ceil;
}

void printVectorToFile(std::vector<double>& vector, const std::string & file_name){
    std::ofstream myfile;
    myfile.open(file_name, std::ios::out | std::ios::app); // output at the end of file
    myfile << "[";
    for (unsigned int i = 0; i < vector.size(); ++i){
        myfile << vector[i];
        if (i < vector.size() - 1)
            myfile << ", ";
    }
    myfile << "]\n";
    myfile.close();
}


void printStringToFile(const std::string & string, const std::string & file_name){
    std::ofstream myfile;
    myfile.open(file_name, std::ios::out | std::ios::app); // output at the end of file
    myfile << string;
    myfile.close();
}

void switchStyle(const std::string & style){
    int code = 0;
    if (style == "black"){
        code = 30;        
    } else if (style == "red"){
        code = 31;
    } else if (style == "green"){
        code = 32;
    } else if (style == "yellow"){
        code = 33;
    } else if (style == "blue"){
        code = 34;
    } else if (style == "magenta"){
        code = 35;
    } else if (style == "cyan"){
        code = 36;
    } else if (style == "white"){
        code = 37;
    } else if (style == "reset"){
        code = 0;
    } else if (style == "bold"){
        code = 1;
    } else if (style == "underline"){
        code = 4;
    } else if (style == "inverse"){
        code = 7;
    }
    std::cout << "\033[" << code << "m";
}

double getSumoCapacity(int green_time){
    int size = sumo_capacity.size();
    double capacity;
    if (green_time >= extra_green + 1){
        if (green_time - extra_green - 1 < size)
            capacity = sumo_capacity[green_time - extra_green - 1];
        else
            capacity = sumo_capacity[size - 1] + (green_time - extra_green - size) * 0.5;  // 1 car per 2 seconds
    } else {
        capacity = 0;
    }
    return capacity;
}

double getDepartureMoment(int k){
    int size = dep_moments.size();
    if (k < 0) {
        std::cout << "incorrect departure index: " << k << "\n";
        throw "Negative departure index\n";
    } else if (k < size){
        return dep_moments[k];
    } else {
        return dep_moments[size - 1] + 2 * (k - size + 1); // 2 seconds for each departure
    }
}


int getArrivalMoment(int k){
    int size = arr_moments.size();
    if (k < size){
        return arr_moments[k];
    } else {
        return arr_moments[size - 1] + 2 * (k - size + 1); // 2 seconds for each departure
    }
}

