#ifndef LAST_ARRIVALS_H
#define LAST_ARRIVALS_H
#include <vector>
#include <map>
#include "state.h"
#include "general.h"
#include "departurechain.h"
#include "tail.h"
// #include "lane.h"

class Lane;
// using KeyIA = std::pair<int, char>;
// using KeyDA = std::pair<Pair, char>;
// using MapIAProb = std::map<KeyIA, double>;

class LastArrivals{
public:
    std::map<KeyDA, MapIAProb> last_arrivals;  // PLA
    std::map<Pair, MapIAProb> la_before_departure;  // PBD
    std::map<KeyDA, std::map<int, MapIAProb>> last_delayed_arrivals; // PLDA  (\bar i' may be <= \bar i!)
    double initialise(int step, std::map<KeyDA, double>& zero_probs, Lane & downsteam_lane);
    void initialiseInTheBeginning(int step, Pair lane_state, Lane & downsteam_lane, bool queue_is_empty, bool is_arrival);
    void updateNewSecond(int step, DepartureChain & departure_chain, const std::map<char, std::map<char, double>>&  ar_transitions,  
                                   const std::map<KeyDA, double> & just_emptied_probs, Lane& downsteam_lane, double real_travel_time);
    bool checkTotalProbabilities(const std::map<KeyDA, double>& zero_probs, DepartureChain & departure_chain);
    bool checkTotalProbabilities(DepartureChain & departure_chain);
    double getTotalProbability();
    void printLastArrivals();
    
    void updateFromTails(std::map<KeyIA, std::vector<int>> & ordered_tails, std::vector<Tail>& tails, 
                         std::vector<double> & unblocked_probabilities_of_tails);
    void removeStates(const std::vector<Pair> & d_keys_to_remove);
};

void updateToNewSecond(int second, MapIAProb & my_map, const std::map<char, double> & b_probs);
double updatePBD(std::map<Pair, MapIAProb> & la_before_departure, int second, int downstream_second, std::map<KeyIA, double> & untransitioned, 
                 const std::map<char, double> & b_tr_probabilities);

bool checkDifferentLastArrivals(std::map<int, LastArrivals> & last_arrivals);
std::map<Pair, double> getNonZeroProbs(const LastArrivals & last_arrivals);

#endif // LAST_ARRIVALS_H
