#ifndef NETWORK_H
#define NETWORK_H

#include <vector>
#include "lane.h"

class RoutingProbability{
public:
    int u_from, u_to;
    double probability;
    RoutingProbability(int u_from, int u_to, double prob) : u_from(u_from), u_to(u_to), probability(prob) {}
};

class Network{
    std::vector<Lane> lanes;
    std::vector<RoutingProbability> routing_probabilities_vector;
    std::map<int, std::map<int, double>> travel_times;
    void updateDepartures(int step);
    int starting_second;
public:
    Network(const std::string & lane_file, const std::string & arrival_rates_file_name, const std::string & routing_file_name, const std::string & green_times_file_name, 
            const std::string & state_file_name); // from a particular (non-zero) state
    Network(const std::string & lane_file, const std::string & arrival_rates_file_name, const std::string & routing_file_name, const std::string & green_times_file_name); // from an empty network
    Network(int duration, int cycle, int green, int small_green, int yellow, int offset, 
            double lamb, double minor_lamb, double factor, double pr_north, double pr_east); // simple hand-made network
    
    double updateOneStep(int step);
    
    void printResultsToFile(int duration, const std::string & file_name);
    void printDistributionsToFile(int duration, const std::string & file_name);
    void printRightDistributionsToFile(int duration, const std::string & file_name);
    double getTotalQueue(int duration);
    double getTotalSquareQueue(int duration);
    
    double run(int duration);
};

#endif // NETWORK_H
