#include <iostream>
#include <algorithm>    // std::swap_ranges
#include <cmath>
#include "queue.h"
#include "general.h"

int TOTAL_QUEUE_COPIES = 0;
int TOTAL_QUEUE_ASSIGNMENTS = 0;
int TOTAL_QUEUE_DEFAULT_CONSTRUCTIONS = 0;
int TOTAL_QUEUE_NORMAL_CONSTRUCTIONS = 0;
int TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS = 0;
int TOTAL_QUEUE_DEPARTURE_CONSTRUCTORS = 0;
int TOTAL_QUEUE_MIX_CONSTRUCTIONS = 0;
int TOTAL_EMPLACE = 0;

Queue::Queue(): capacity(-1), queue(0){
//     queue.resize(1);
    start = 0;
    end = 0;
    ++TOTAL_QUEUE_DEFAULT_CONSTRUCTIONS;
//     std::cout << "Default constructor is used\n";
}


Queue::Queue(int capacity): capacity(capacity){
    queue.resize(capacity + 1);
    start = 0;
    end = 0;
    ++TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS;
}

void Queue::emplace(int new_capacity){
    capacity = new_capacity;
    queue.resize(capacity + 1);
    start = 0;
    end = 0;
    ++TOTAL_EMPLACE;
}

bool doesEmplaceMakeSense(int new_capacity) { return (new_capacity > 0); } // return true if it makes sense to add this queue
bool doesEmplaceMakeSense(int new_capacity, double probability) { return (probability > almost_zero) && (new_capacity > 0); }

Queue::Queue(int capacity, int new_start, int new_end): capacity(capacity), start(new_start), end(new_end){
    queue.resize(capacity + 1, 0);
    ++TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS;
}


void Queue::emplace(int new_capacity, int new_start, int new_end){
    capacity = new_capacity;
    queue.resize(capacity + 1, 0);
    start = new_start;
    end = new_end;
    ++TOTAL_EMPLACE;
}

bool doesEmplaceMakeSense(int new_capacity, int new_start, int new_end){
    return (new_end > new_start) && (new_capacity > 0);
}


Queue::Queue(int capacity, double probability): capacity(capacity){
    queue.resize(capacity + 1);
    queue[0] = probability;
    start = 0;
    end = 1;
    ++TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS;
}


void Queue::emplace(int new_capacity, double probability){
    capacity = new_capacity;
    queue.resize(capacity + 1);
    queue[0] = probability;
    start = 0;
    end = 1;
    ++TOTAL_EMPLACE;
}


Queue::Queue(int capacity, int queue_length, double probability) : capacity(capacity){
    queue.resize(capacity + 1);
    queue[queue_length] = probability;
    start = queue_length;
    end = queue_length + 1;
    ++TOTAL_QUEUE_SIMPLE_CONSTRUCTIONS;
}

void Queue::emplace(int new_capacity, int queue_length, double probability){
    capacity = new_capacity;
    queue.resize(capacity + 1);
    queue[queue_length] = probability;
    start = queue_length;
    end = queue_length + 1;
    ++TOTAL_EMPLACE;
}

bool doesEmplaceMakeSense(int new_capacity, int queue_length, double probability){
    return (probability > almost_zero) && (new_capacity >= queue_length);
}


Queue::Queue(std::vector<double>& probabilities): capacity(probabilities.size() - 1), queue(probabilities) {
//     queue = std::vector<double>(probabilities);
    start = 0;
    end = capacity + 1;
}

Queue::Queue(std::vector<double>& probabilities, int start, int end): capacity(probabilities.size() - 1), start(start), end(end) {
    queue.resize(capacity + 1);
//     if (start == 0){
//         std::cout << "A queue was made with start at 0\n";
//         std::cout << end << " " << probabilities << "\n";
//         double total = 0;
//         for (int i = start; i < end; ++i){
//             total += probabilities[i];
//         }
//         if (total > small_eps && total < 0.001 && end > 5) {
//             std::cout << total << "\n";
//             std::cout << "Essential queue\n";
//         }
//     }
//      
//     std::cout << "in normal constructor\n";
//     std::cout << start << " " << end << "\n";
//     std::cout << capacity << "\n";
    
    std::copy(probabilities.begin() + start, probabilities.begin() + end, queue.begin() + start);
//     std::cout << "copied\n";
    ++TOTAL_QUEUE_NORMAL_CONSTRUCTIONS;
}

Queue::Queue(const Queue& another, int start, int end): capacity(another.capacity), start(start), end(end) {
    queue.resize(capacity + 1);
    std::copy(another.queue.begin() + start, another.queue.begin() + end, queue.begin() + start);
    ++TOTAL_QUEUE_NORMAL_CONSTRUCTIONS;
}

void Queue::emplace(const Queue& another, int new_start, int new_end){
    start = new_start;
    end = new_end;
    capacity = another.capacity;
    queue.resize(capacity + 1);
    std::copy(another.queue.begin() + start, another.queue.begin() + end, queue.begin() + start);
    ++TOTAL_EMPLACE;
}

// template<>
bool doesEmplaceMakeSense(const Queue& another, int new_start, int new_end){
    if (new_start >= new_end)
        return false;
    double total = 0;
    for (int i = new_start; i < new_end; ++i){
        total += another[i];
    }
    return (total > almost_zero);
}




Queue::Queue(const Queue& another): capacity(another.capacity), queue(capacity + 1){
//     std::cout << "in queue copy constructor \n";
//     queue.resize(capacity + 1);
//     std::cout << "Copy constructor is used\n";
    ++TOTAL_QUEUE_COPIES;
    start = another.start;
    end = std::max(another.end, start);  // if the end is smaller than start a segmentation fault in the next line occurs
    std::copy(another.queue.begin() + start, another.queue.begin() + end, queue.begin() + start);
}

Queue::Queue(const Queue& another, double coefficient): capacity(another.capacity), queue(capacity + 1){
//     queue.resize(capacity + 1);
    ++TOTAL_QUEUE_COPIES;
    start = another.start;
    end = std::max(another.end, start);
    std::transform(another.queue.begin() + start, another.queue.begin() + end, queue.begin() + start,
                   [coefficient](double x) { return x * coefficient; });
}

Queue::Queue(int capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, 
             bool is_free_departure_possible): capacity(capacity){
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    mixQueues(queues, coefficients, coefficients_full, is_free_departure_possible);
    end = std::max(start, end);
    ++TOTAL_QUEUE_MIX_CONSTRUCTIONS;
}

void Queue::mixQueues(const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, bool is_free_departure_possible){
    int other_start, other_end, shift;
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        if (queues.hasKey(a_ind) && !queues[a_ind].hasZeroProbability()){ // not an empty queue (empty in the sense start >= end)
            shift = (a_ind == KEY_A) ? 1 : 0;
            other_start = queues[a_ind].start;
            other_end = queues[a_ind].end;
            if (other_start > 0 && !is_free_departure_possible)
                start = std::min(start, other_start + shift);
            else
                start = std::min(start, other_start);                
            end = std::max(end, std::min(other_end + shift, capacity + 1));
            for (int i = other_start; i < other_end; ++i){
                if (i < capacity - shift){
                    if (coefficients.hasKey(a_ind)){
                        if (i > 0 || !is_free_departure_possible)
                            queue[i + shift] += coefficients[a_ind] * queues[a_ind][i];
                        else // i = 0 and free departure is possible
                            queue[0] += coefficients[a_ind] * queues[a_ind][i];
                    }
                } else {
                    if (coefficients_full.hasKey(a_ind))
                        queue[capacity] += coefficients_full[a_ind] * queues[a_ind][i];
                }
            }
        }
    }
}


inline void min_me(int & me, int other){
    // me = min(me, other)
    if (me > other)
        me = other;
}

inline void max_me(int & me, int other){
    // me = max(me, other)
    if (me < other)
        me = other;
}


void Queue::mixQueues(const ABPMap<Queue>& queues, const std::map<char,double>& coefficients, 
                      const std::map<char,double>& coefficients_full, bool is_free_departure_possible){
    int other_start, other_end, shift;
    char a_key;
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        if (queues.hasKey(a_ind) && !queues[a_ind].hasZeroProbability()){ // not an empty queue (empty in the sense start >= end)
            a_key = getAKey(a_ind);
            shift = (a_ind == KEY_A) ? 1 : 0;
            other_end = queues[a_ind].end;
            if (other_end + shift >= capacity && isInMap(coefficients_full, a_key)) {
                max_me(end, capacity + 1);
                queue[capacity] += coefficients_full.at(a_key) * queues[a_ind][capacity - shift];
                if (shift == 1 && other_end == capacity + 1){
                    queue[capacity] += coefficients_full.at(a_key) * queues[a_ind][capacity];
                }
            }
            if (isInMap(coefficients, a_key)){
                other_start = queues[a_ind].start;
                if (other_start == 0 && is_free_departure_possible) {
                    min_me(start, other_start);
                    queue[0] += coefficients.at(a_key) * queues[a_ind][0];
                    other_start = 1;
                } else {
                    min_me(start, other_start + shift);
                }
                min_me(other_end, capacity - shift);
                max_me(end, other_end + shift);
                for (int i = other_start; i < other_end; ++i){
                    queue[i + shift] += coefficients.at(a_key) * queues[a_ind][i];
                }
            }
        }
    }
}

/*
void Queue::emplace(int new_capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, 
                    const ABPMap<double>& coefficients_full, bool is_free_departure_possible){
    capacity = new_capacity;
    queue.clear();
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    mixQueues(queues, coefficients, coefficients_full, is_free_departure_possible);
    end = std::max(start, end);
    ++TOTAL_EMPLACE;
}*/


void Queue::emplace(int new_capacity, const ABPMap<Queue>& queues, const std::map<char,double>& coefficients, 
                    const std::map<char,double>& coefficients_full, bool is_free_departure_possible){
    capacity = new_capacity;
    queue.clear();
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    mixQueues(queues, coefficients, coefficients_full, is_free_departure_possible);
    end = std::max(start, end);
    ++TOTAL_EMPLACE;
}

bool doesEmplaceMakeSense(int new_capacity, const ABPMap<Queue>& queues, const std::map<char, double>& coefficients,
                                 const std::map<char, double>& coefficients_full, bool is_free_departure_possible){
//     int new_start = new_capacity + 1, new_end = 0;
    int other_start, other_end, shift;
    char a_key;
    for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
        if (queues.hasKey(a_ind) && !queues[a_ind].hasZeroProbability()){
            a_key = getAKey(a_ind);
            other_start = queues[a_ind].getStart();
            shift = (a_key == 'a') ? 1: 0;
            if (isInMap(coefficients, a_key) && coefficients.at(a_key) > almost_zero){
                other_end = queues[a_ind].getEnd();
                min_me(other_end, new_capacity - shift);
                if (other_start < other_end && queues[a_ind].isPartialProbabilitySignificant(other_start, other_end, coefficients.at(a_key))){
                    return true;
                }
            }
            if (isInMap(coefficients_full, a_key) && coefficients_full.at(a_key) > almost_zero){
                other_end = queues[a_ind].getEnd();
//                 std::cout << "checking coef full " << a_ind << " aka '" << a_key << "', other start, end are " << other_start 
//                           << " " << other_end << ", capacity is " << new_capacity;
                if (other_start <= new_capacity - shift && other_end >= new_capacity - shift + 1 && 
                    queues[a_ind].isPartialProbabilitySignificant(new_capacity - shift, other_end, coefficients.at(a_key))){
                    return true;
                }
            }
        }
    }
    if (is_free_departure_possible) {} // do nothing just suppresses the warning of unused variable
//     std::cout << "Emplace does not make sense!\n";
    return false;
}



// Queue::Queue(int capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full): capacity(capacity){
//     queue.resize(capacity + 1, 0);
//     start = capacity + 1;
//     end = 0;
//     int other_start, other_end;
//     for (int a_ind = 0; a_ind < NUMBER_OF_KEYS; ++a_ind){
//         if (queues.hasKey(a_ind) && !queues[a_ind].hasZeroProbability()){ // not an empty queue (empty in the sense start >= end)
//             other_start = queues[a_ind].start;
//             other_end = queues[a_ind].end;
//             start = std::min(start, other_start);
//             end = std::max(end, other_end);
//             for (int i = other_start; i < other_end; ++i){
//                 if (i != capacity){
//                     if (coefficients.hasKey(a_ind))
//                         queue[i] += coefficients[a_ind] * queues[a_ind][i];
//                 } else {
//                     if (coefficients_full.hasKey(a_ind))
//                         queue[i] += coefficients_full[a_ind] * queues[a_ind][i];
//                 }
//             }
//         }
//     }
//     end = std::max(start, end);
//     ++TOTAL_QUEUE_MIX_CONSTRUCTIONS;
// }

Queue::Queue(int capacity, const Queue& queue1, const Queue& queue2, double coef1, double coef2): capacity(capacity){
    queue.resize(capacity + 1, 0);
    if (queue1.start < queue1.end){
        std::transform(queue1.queue.begin() + queue1.start, queue1.queue.begin() + queue1.end, queue.begin() + queue1.start,
                   [coef1](double x) { return x * coef1; });
        start = queue1.start;
        end = queue1.end;
    } else {
        start = 0;
        end = 0;
    }
    if (queue2.start < queue2.end){
        for (int i = queue2.start; i < queue2.end; ++i)
            queue[i] += queue2[i] * coef2;
//         std::transform(queue2.queue.begin() + queue2.start, queue2.queue.begin() + queue2.end, queue.begin() + queue2.start,
//                    [coef2](double x) { return x * coef2; });  /// this shouldn't work
        if (start < end){
            start = std::min(start, queue2.start);
            end = std::max(end, queue2.end);
        } else {
            start = queue2.start;
            end = queue2.end;
        }
    } 
    ++TOTAL_QUEUE_MIX_CONSTRUCTIONS;
}

Queue::Queue(int capacity, const Queue& another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue = false*/): capacity(capacity){
    // this is equivalent to Queue(capacity, another.afterNoDeparture (m_second - 1), another.afterDeparture, coef_no_dep, coef_dep) 
    // but does not construct extra queues
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    addDepartureQueue(another, m_second, coef_no_dep, coef_dep/*, is_split_queue*/);
    ++TOTAL_QUEUE_DEPARTURE_CONSTRUCTORS;
}

void Queue::emplace(int new_capacity, const Queue& another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue = false*/){
    capacity = new_capacity;
    queue.clear();
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    addDepartureQueue(another, m_second, coef_no_dep, coef_dep);
    ++TOTAL_EMPLACE;
}

bool doesEmplaceMakeSense(int new_capacity, const Queue& another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue*/){
    int from = another.getStart(); // this is if zero departure prob
    int to = std::max(m_second, from);
    if (coef_dep >= almost_zero) {
        if (from != 0){
            --from; // departure
        }
    } else if (coef_no_dep < almost_zero) {
        return false; // small probs
    }
    if (from >= to)
        return false;
    double total = 0;
    for (int i = from; i < to; ++ i)
        total += (coef_no_dep + coef_dep) * another[i];
    return total > almost_zero && new_capacity > 0; 
}



void Queue::addDepartureQueue(const Queue& another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue = false*/){
    int to = std::max(m_second, another.start);
    int from = another.start; // this is if zero departure prob
    if (coef_no_dep > 0){
        for (int i = from; i < to; ++ i)
            queue[i] += coef_no_dep * another[i];
    }
    if (coef_dep > 0) {
        if (another.start == 0){
//         } else { // another.start == 0  this shouldn't happen
            queue[0] += another[0] * coef_dep;
        }
        max_me(from, 1);
//         from = std::max(another.start, is_split_queue ? 2: 1); // do not add emptied probability
        for (int i = from; i < to; ++ i){
            queue[i - 1] += another[i] * coef_dep;
        }
        from = (another.start != 0) ? from - 1: 0; // because queue[i-1] depends on  another[i]
    }
    start = std::min(start, from);
    end = std::max(std::max(end, to), start);
}



Queue::Queue(int capacity, const std::vector<Queue>& queues, const std::vector<double>& coefficients): capacity(capacity){
    queue.resize(capacity + 1, 0);
    start = capacity + 1;
    end = 0;
    for (unsigned int j = 0; j < queues.size(); ++j){
        if (queues[j].start < queues[j].end){
            start = std::min(start, queues[j].start);
            end = std::max(end, queues[j].end);
        }
    }
    for (int i = start; i < end; ++i){
        queue[i] = 0.;
        for (unsigned int j = 0; j < queues.size(); ++j){
            if ((queues.at(j).start <= i) && (queues.at(j).end > i))
                queue[i] += queues.at(j)[i] * coefficients[j];
        }
    }
    ++TOTAL_QUEUE_MIX_CONSTRUCTIONS;
}



Queue & Queue::operator=(const Queue& another){
//     std::cout << "Assignment is used\n";
    ++TOTAL_QUEUE_ASSIGNMENTS;
    capacity = another.capacity;
    start = another.start;
    end = std::max(another.end, start);
    queue.resize(capacity + 1);
    std::copy(another.queue.begin() + start, another.queue.begin() + end, queue.begin() + start);
    return *this;
}


void Queue::printQueue() const {
    std::cout << "Printing queue of capacity " << capacity << ", it has start = " << start << " and end = " << end << ", ";
    std::cout << "total probability is " << getTotalProbability() << ": ";
    for (int i = 0; i < capacity + 1; ++i){
        double x = ((i >= start) && (i < end)) ? queue[i] : 0;  // queue probabilities outside [start, end) are not properly defined
        std::cout << " " << x;
    }
    std::cout << "\n";
}

double Queue::operator[](int i) const{
    return ((i >= start) && (i < end)) ? queue[i] : 0;
}


Queue Queue::afterArrival(bool is_free_departure_possible) const{
    std::vector<double> new_probabilities(capacity + 1);
    int new_start = start, new_end = start;
//     new_probabilities.resize(capacity + 1);
    if (start < end){
        int shift_from = is_free_departure_possible ? 1 : 0;
        shift_from = std::max(start, shift_from);
        std::copy(queue.begin() + shift_from, queue.begin() + std::max(shift_from, std::min(capacity, end)), 
                  new_probabilities.begin() + 1 + shift_from);
        new_start = start + 1;
        new_end = end + 1;
        if (end == capacity + 1){
            new_probabilities[capacity] += queue[capacity];
            new_end = end;
            if (start == capacity)
                new_start = start;
        }
        if (is_free_departure_possible && start == 0){
            new_probabilities[0] = queue[0];
            new_start = start;
        }
    }
//     switchStyle("blue");
//     std::cout << "before arrival: " << start << " " << end << std::endl;
//     std::cout << "after arrival: " << new_start << " " << new_end << std::endl;
//     switchStyle("reset");
    return Queue(new_probabilities, new_start, std::max(new_end, new_start));
}


Queue Queue::afterDeparture(int n){
    // do not! set probabilities queue[i] to 0 for i <= n
    // return the part of queue[0] + queue[1], ..., queue[n] as a new queue of the same capacity
    std::vector<double> new_probabilities(capacity + 1, 0.);
//     new_probabilities.resize(capacity + 1, 0.);
    int swap_from = std::max(1, start);
    int swap_to = std::max(swap_from, std::min(n + 1, end));
    int new_end = swap_to - 1;
    std::copy(queue.begin() + swap_from, queue.begin() + swap_to, new_probabilities.begin() + swap_from - 1);
    if (start == 0){
        new_probabilities[0] += queue[0];
//         queue[0] = 0;
        new_end = std::max(new_end, 1);
    }
    int new_start = swap_from - 1; 
//     start = std::max(start, n + 1);
//     end = std::max(start, end);
    return Queue(new_probabilities, new_start, new_end);
}

Queue Queue::afterNoDeparture(int n){
    // set probabilities queue[i] to 0 for i <= n
    // return the part of queue[0], ..., queue[n] as a new queue of the same capacity
    std::vector<double> new_probabilities(capacity + 1, 0.);
//     new_probabilities.resize(capacity + 1, 0.);
    int new_start = start;
    int new_end = std::max(start, std::min(n + 1, end));
    std::swap_ranges(queue.begin() + start, queue.begin() + new_end, new_probabilities.begin() + start);
    start = std::max(start, n + 1);
    end = std::max(start, end);
    return Queue(new_probabilities, new_start, new_end);
}

double stopLineSecond(int second, int m){
    return second - std::max(m - 1, 0) * bumper_to_bumper * (1/free_flow_speed + 1/ wave_speed);
}

int minNotAffectedQueueLength(int second, int downstream_second, int capacity){
    // Return minimum such m_second that s(second, m_second) <= downstream_second
    // second - (m_second-1)^+ * bumper_to_bumper * (1/free_flow_speed + 1/ wave_speed) <= downstream_second
    if (second <= downstream_second)
        return 0;
    int m_second = std::ceil((second - downstream_second - bumper_to_bumper/free_flow_speed) / (bumper_to_bumper * (1 / free_flow_speed + 1 / wave_speed))) + 1; // first not-affected vehicle
    m_second = std::min(m_second, capacity + 1);
    return m_second;
}


Queue Queue::forTail ( int step, int i_from, bool is_from_a, bool is_block, double coefficient ){
    std::vector<double> new_probabilities(capacity + 1);
    int new_start, new_end;
    int m_second = minNotAffectedQueueLength(i_from, step - 1, capacity);
    if ((m_second > capacity) || ((m_second > capacity - 1) && is_from_a)){
        throw "The lane seems to be too short, forTail queue starts after blocking probabilities!\n";
    }
    new_start = std::max(m_second, start);
    new_end = end;
    if (is_block) {
        new_start = std::max(new_start, (is_from_a) ? capacity - 1:  capacity);
    } else {
        new_end = std::min(new_end, (is_from_a) ? capacity - 1:  capacity);
    }
//     new_probabilities.resize(capacity + 1);
    for (int i = new_start; i < new_end; ++i) {
        new_probabilities[i] = queue[i] * coefficient;
        queue[i] *= (1 - coefficient); // remove from the (untransitioned) queue
    }
//     std::cout << new_start << "\n";
    return Queue(new_probabilities, new_start, std::max(new_start, new_end));
}

void Queue::removeTailPart(int new_start, int new_end, double coefficient){
    for (int i = new_start; i < new_end; ++i) {
        queue[i] *= (1 - coefficient); // remove from the (untransitioned) queue
    }
}

bool doesEmplaceMakeSense(const Queue & another, int m_second, bool is_from_a, bool is_block, double coefficient){ // forTail emplace
    int new_capacity = another.getCapacity();
    int start = std::max(m_second, another.getStart());
    int end = another.getEnd();
    if (is_block) {
        max_me(start, (is_from_a) ? new_capacity - 1:  new_capacity);
    } else {
        min_me(end, (is_from_a) ? new_capacity - 1:  new_capacity);
    }
//     std::cout << start << ", " << end << "\n";
//     std::cout << coefficient  << " * " << coefficient * another.getPartialProbability(start, end)<< "\n";
    return (start < end && another.isPartialProbabilitySignificant(start, end, coefficient));
}


void Queue::emplace(const Queue & another, int m_second, bool is_from_a, bool is_block, double coefficient){ // forTail emplace
    capacity = another.getCapacity();
    queue.resize(capacity + 1);
    start = std::max(m_second, another.getStart());
    end = another.getEnd();
    if (is_block) {
        max_me(start, (is_from_a) ? capacity - 1:  capacity);
    } else {
        min_me(end, (is_from_a) ? capacity - 1:  capacity);
    }
//     // it is NOT assumed that doesEmplaceMakeSense was used and thus start and end are NOT YET computed!
    for (int i = start; i < end; ++i) {
        queue[i] = another[i] * coefficient;
//         another.queue[i] *= (1 - coefficient); // remove from the (untransitioned) queue // TODO: remove
    }
    ++TOTAL_EMPLACE;
}

Queue Queue::getMiddlePart (int second, int downstream_second){
    // get queue[m] for m_second > m >= m_{second-1}, where s(second +1, m_second) <= downstream_second ad m_second is the smallest.
    // set queue[m] = 0 for such m
    int m_up = minNotAffectedQueueLength(second, downstream_second-1, capacity);
    int m_down = minNotAffectedQueueLength(second, downstream_second, capacity);
    int new_start = std::max(start, m_down);
    int new_end = std::max(new_start, std::min(end, m_up));
    std::vector<double> new_probabilities(capacity + 1, 0);
//     new_probabilities.resize(capacity + 1, 0);
    std::swap_ranges(queue.begin() + new_start, queue.begin() + new_end, new_probabilities.begin() + new_start);
    if (start >= new_start) // means ==
        start = new_end; // thus between new_start and new_end probabilities are zeros
    if (end <= new_end)  // means == 
        end = new_start;
    end = std::max(start, end);
    return Queue(new_probabilities, new_start, new_end);    
}

void swapQueueParts(Queue & queue1, Queue & queue2, int from, int to){
    std::swap_ranges(queue1.queue.begin() + from, queue1.queue.begin() + to, queue2.queue.begin() + from);
}

// Queue::Queue(const Queue& another, int new_start, int new_end) : capacity(another.capacity), start(new_start), end(new_end){
//     // =getMiddlePart
//     queue.resize(capacity + 1, 0);
//     std::swap_ranges(another.queue.begin() + new_start, another.queue.begin() + new_end, queue.begin() + new_start);
// //     if (another.start >= new_start)
// //         another.start = new_end;
// //     if (another.end <= new_end)
// //         another.end = new_start;
// //     another.end = std::max(another.start, another.end);
// //     switchStyle("magenta");
// //     printQueue();
// //     another.printQueue();
// //     switchStyle("reset");
// }


void Queue::getStartEndMiddle(int m_down, int m_up, int& new_start, int& new_end){
//     int m_up = minNotAffectedQueueLength(second, downstream_second-1, capacity);
//     int m_down = minNotAffectedQueueLength(second, downstream_second, capacity);
    new_start = std::max(start, m_down);
    new_end = std::max(new_start, std::min(end, m_up));
}

Queue Queue::getUpstreamSide (int second, int downstream_second, bool to_delete /* = false*/){
    // get queue[m] for m >= m_second, where s(second +1, m_second) <= downstream_second ad m_second is the smallest.
    int m_second = minNotAffectedQueueLength(second, downstream_second - 1, capacity);
    int new_start = std::max(start, m_second);
    int old_end = end;
//     std::vector<double> new_probabilities;
//     std::copy(queue.begin() + new_start, queue.begin() + end, new_probabilities.begin() + new_start);
//     std::cout << new_start << " " << end << std::endl;
    if (to_delete){
        end = new_start;
        start = std::min(start, end);
    }
    return Queue(queue, new_start, std::max(new_start, old_end));    
}


Queue Queue::getDownstreamPart(int second, int downstream_second){
    int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
    return Queue(queue, std::min(start, m_second), std::min(std::max(start, end), m_second));
}

void Queue::getStartEndDownstream(int m_second, int& new_start, int& new_end){
    new_start = start;
    new_end = std::max(std::min(m_second, end), new_start);
}



void Queue::removeMiddlePart ( int second, int downstream_second ){
    // "Set" to zero everything with  downstream_second - 1 < s(second, m) <= downstream_second
    int m_up = minNotAffectedQueueLength(second, downstream_second-1, capacity);
    int m_down = minNotAffectedQueueLength(second, downstream_second, capacity);
    if (start >= m_down)
        start = std::max(m_up, start);
    if (end <= m_up)
        end = std::min(m_down, end);
    if ((start < m_down) && (end > m_up)){
        for (int i = m_down; i < m_up; ++i)
            queue[i] = 0;
    }
}

void Queue::removeMiddlePartKnowingMs(int m_down, int m_up, bool make_zero /* = true*/){
    if (start >= m_down)
        start = std::max(m_up, start);
    if (end <= m_up)
        end = std::min(m_down, end);
    if (make_zero && (start < m_down) && (end > m_up)){
        for (int i = m_down; i < m_up; ++i)
            queue[i] = 0;
    }
}



bool Queue::removeDownstreamPart ( int second, int downstream_second ){
    // "Set" to zero everything with s(second, m) > downstream_second
    int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
    start = std::max(start, m_second);
    end = std::max(start, end);
    return (start >= end); // return true if the queue is empty after removing
}

bool Queue::removeDownstreamPart(int m_second){
    start = std::max(start, m_second);
    end = std::max(start, end);
    return (start >= end);
}



void Queue::addQueue(const Queue& another){
    if (another.start < another.end){
        if (start >= end){
            start = another.start;
            end = start;
        }
        for (int i = another.start; i < start; ++ i)
            queue[i] = 0;
        for (int i = end; i < another.end; ++ i)
            queue[i] = 0;
        for (int i = another.start; i < another.end; ++i){
            queue[i] += another.queue[i];
        }
        start = std::min(start, another.start);
        end = std::max(end, another.end);
    }
}

void Queue::addUpstreamAndMiddleQueueByCoef(const Queue& another, double coefficient, double coefficient_full, bool is_a_queue, int second, int downstream_second){
    int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
    int another_start_alt = std::max(m_second, another.start);
//     std::cout << "adding upstream and middle\n";
//     std::cout << "up from " << m_second << "\n";
//     std::cout << start << " " << end << " " << another_start_alt << " " << another.end << "\n";
//     std::cout << coefficient << " " << coefficient_full << "\n";
    if (another_start_alt < another.end){
        if (start >= end){
            start = another_start_alt;
            end = another_start_alt;
        }
        int full_index = (is_a_queue) ? capacity - 1: capacity;
        for (int i = another_start_alt; i < start; ++i){
            queue[i] = 0;
        }
        for (int i = end; i < another.end; ++i){
            queue[i] = 0;
        }
        for (int i = another_start_alt; i < std::min(another.end, full_index); ++i){
            queue[i] += coefficient * another[i];
        }
        if (another.end > full_index){ // other coefficient for full queue
            for (int i = full_index; i < std::min(another.end, capacity + 1); ++i){
                queue[i] += coefficient_full * another[i];
            }
        }
        start = std::min(start, another_start_alt);
        end = std::max(end, another.end);
    }
}



void Queue::addDownstreamPartOfQueue(const Queue& another, int second, int downstream_second){
    int m_second = minNotAffectedQueueLength(second, downstream_second, capacity);
    int another_end_alt = std::min(another.end, m_second);
    if (another.start < another_end_alt){
        if (start >= end){
            start = another.start;
            end = start;
        }
        for (int i = another.start; i < start; ++ i)
            queue[i] = 0;
        for (int i = end; i < another_end_alt; ++ i)
            queue[i] = 0;
        for (int i = another.start; i < another_end_alt; ++i){
            queue[i] += another.queue[i];
        }
        start = std::min(start, another.start);
        end = std::max(end, another_end_alt);
    }
}

// bool Queue::isPartialProbabilitySignificant(int from, double coef/* = 1.*/){
//     double sum = 0;
//     max_me(from, start);
//     for (int i = from; i < end; ++i){
//         sum += queue[i] * coef;
//         if (sum > almost_zero)
//             return true;
//     }
//     return false;
// }

bool Queue::isPartialProbabilitySignificant(int from, int to, double coef/* = 1.*/) const{
    double sum = 0;
    max_me(from, start);
    min_me(to, end);
    for (int i = from; i < to; ++i){
        sum += queue[i] * coef;
        if (sum > almost_zero)
            return true;
    }
    return false;
}



// double Queue::getPartialProbability(int from){
//     double sum = 0;
//     for (int i = from; i < end; ++i)
//         sum += this->operator[](i);
//     return sum;
// }
// 
// double Queue::getPartialProbability(int from, int to) const{
//     double sum = 0;
//     for (int i = from; i < to; ++i)
//         sum += this->operator[](i);
//     return sum;
// }


double Queue::getTotalProbability() const {
    double sum = 0;
    for (int i = start; i < end; ++i)
        sum += queue[i];
    return sum;
}


double Queue::getAlmostFullProbability(bool is_a_queue) const {
    double sum = 0;
    int from_which = (is_a_queue) ? capacity - 1: capacity;
    for (int i = std::max(start, from_which); i < end; ++i)
        sum += queue[i];
    return sum;
}


double Queue::getExpectedQueueSquare(int start_from) const {
    double result = 0.;
//     std::cout << "in\n";
//     std::cout << start << " " << end << "\n";
    for (int i = std::max(1, start); i < end; ++i) {
//         std::cout << i << "\n";
        result += (i + start_from) * (i + start_from) * queue[i];
    }
//     std::cout << "out\n";
    return result;
}

double Queue::getExpectedQueueLength() const {
    double result = 0.;
    for (int i = std::max(1, start); i < end; ++i)
        result += i * queue[i];
    return result;
}

double Queue::getExpectedQueueLengthInMeters() const {
    double result = 0.;
    for (int i = std::max(1, start); i < end; ++i)
        result += (std::max(0, i - 1) * bumper_to_bumper + to_stop_line) * queue[i];
    return result;
}

bool Queue::hasZeroProbability() const {
    return (start >= end);
}


bool Queue::hasAlmostZeroProbability() const {
    return (getTotalProbability() < almost_zero);
}

void Queue::setProbability(int i, double value){ 
    queue[i] = value; 
    if (start < end){
        if (value > 0){
            start = std::min(start, i); 
            end = std::max(end, i + 1); 
        } else {
            if (start == i)
                start += 1;
            if (end == i + 1)
                end -= 1;
        }
    } else if (value > 0){
        start = i;
        end = i + 1;
    }
    end = std::max(start, end);
}

void Queue::addProbability(int i, double value){
    setProbability(i, value + this->operator[](i));
}




Queue mixQueues(const std::vector<Queue> & queues, const std::vector<double> & coefficients){
    int capacity = queues[0].getCapacity();
    std::vector<double> new_probabilities(capacity + 1);
//     new_probabilities.resize(capacity + 1);
    int new_start = capacity + 1, new_end = 0;
    for (unsigned int j = 0; j < queues.size(); ++j){
        if (queues[j].start < queues[j].end){
            new_start = std::min(new_start, queues[j].start);
            new_end = std::max(new_end, queues[j].end);
        }
    }
    for (int i = new_start; i < new_end; ++i){
        new_probabilities[i] = 0.;
        for (unsigned int j = 0; j < queues.size(); ++j){
            if ((queues.at(j).start <= i) && (queues.at(j).end > i))
                new_probabilities[i] += queues.at(j)[i] * coefficients[j];
        }
    }
    return Queue(new_probabilities, new_start, std::max(new_end, new_start));
}

Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients){
    std::vector<double> new_probabilities(capacity + 1);
//     new_probabilities.resize(capacity + 1);
    int new_start = capacity + 1, new_end = 0;
    for (auto const& x : queues){
        if (coefficients.find(x.first) != coefficients.end()){
            new_start = std::min(new_start, queues.at(x.first).start);
            new_end = std::max(new_end, queues.at(x.first).end);
        }
    }
    for (int i = new_start; i < new_end; ++i){
        new_probabilities[i] = 0.;
        for (auto const& x : coefficients){
            if ((queues.find(x.first) != queues.end()) && (queues.at(x.first).start <= i) && (queues.at(x.first).end > i))
                new_probabilities[i] += queues.at(x.first)[i] * coefficients.at(x.first);
        }
    }
    return Queue(new_probabilities, new_start, std::max(new_start, new_end));
}


Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients, 
                const std::map<char, double> & coefficients_full){
    std::vector<double> new_probabilities (capacity + 1);
//     new_probabilities.resize(capacity + 1);
    int new_start = capacity + 1, new_end = 0;
    for (auto const& x : queues){
        if (isInMap(coefficients, x.first) || isInMap(coefficients_full, x.first)){
            if (queues.at(x.first).start < queues.at(x.first).end){
                new_start = std::min(new_start, queues.at(x.first).start);
                new_end = std::max(new_end, queues.at(x.first).end);
            }
        }
    }
    for (int i = new_start; i < std::min(new_end, capacity); ++i){
        new_probabilities[i] = 0.;
        for (auto const& x : coefficients){
            if ((queues.find(x.first) != queues.end()) && (queues.at(x.first).start <= i) && (queues.at(x.first).end > i)){
                new_probabilities[i] += queues.at(x.first)[i] * coefficients.at(x.first);
            }
        }
    }
    if (new_end > capacity){ // other transitions for full queues
        int i = capacity;
        new_probabilities[i] = 0.;
        for (auto const& x : coefficients_full){
            if ((queues.find(x.first) != queues.end()) && (queues.at(x.first).start <= i) && (queues.at(x.first).end > i)){
                new_probabilities[i] += queues.at(x.first)[i] * coefficients_full.at(x.first);
            }
        }
    }
    return Queue(new_probabilities, new_start, std::max(new_start, new_end));
}


void Queue::renormaliseQueue(double divide_by){
    for (int i = start; i < end; ++i){
        queue[i] /= divide_by; 
    }
}


