#ifndef QUEUE_H
#define QUEUE_H
#include <vector>
#include <map>
#include "state.h"
#include "abpmap.h"

template<typename... Args> bool doesEmplaceMakeSense(const Args &... args);


class Queue{
    int capacity; // lane capacity = max number of cars at the lane
    int start, end;
    std::vector<double> queue; 
public:
    Queue();
    explicit Queue(int capacity);
    Queue(int capacity, double probability);
    Queue(int capacity, int queue_length, double probability);
    Queue(int capacity, int new_start, int new_end);
    explicit Queue(std::vector<double>& probabilities);
    Queue(std::vector<double>& probabilities, int start, int end);
    Queue(const Queue& another); // copy constructor
    Queue(const Queue& another, double coefficient);
    Queue(const Queue& another, int start, int end); // copy part of another queue
    Queue(int capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, bool is_free_departure_possible);
//     Queue(int capacity, ABPMap<Queue>& queues, ABPMap<double>& coefficients, ABPMap<double>& coefficients_full);
    Queue(int capacity, const std::vector<Queue> & queues, const std::vector<double> & coefficients);
    Queue(int capacity, const Queue & queue1, const Queue & queue2, double coef1, double coef2);
    Queue(int capacity, const Queue & another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue=false*/);
    
    void emplace(int new_capacity);
//     bool doesEmplaceMakeSense(int new_capacity) { return (new_capacity > 0); } // return true if it makes sense to add this queue
    void emplace(int new_capacity, double probability);
    void emplace(int new_capacity, int queue_length, double probability);
//     bool doesEmplaceMakeSense(int new_capacity, double probability) { return (probability > almost_zero) && (new_capacity > 0); }
    void emplace(int new_capacity, const Queue & another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue=false*/);
//     bool doesEmplaceMakeSense(int new_capacity, const Queue & another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue=false*/);
//     void emplace(int new_capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, bool is_free_departure_possible);
//     bool doesEmplaceMakeSense(int new_capacity, const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, bool is_free_departure_possible) {return true; // for now} 
    void emplace(int new_capacity, const ABPMap<Queue>& queues, const std::map<char,double>& coefficients, const std::map<char,double>& coefficients_full, bool is_free_departure_possible);
//     bool doesEmplaceMakeSense(int new_capacity, const ABPMap<Queue>& queues, const std::map<char,double>& coefficients, const std::map<char,double>& coefficients_full, bool is_free_departure_possible);
    void emplace(int new_capacity, int new_start, int new_end);
//     bool doesEmplaceMakeSense(int new_capacity, int new_start, int new_end);
    void emplace(const Queue& another, int start, int end); // copy part of another queue
//     bool doesEmplaceMakeSense(const Queue& another, int start, int end); // copy part of another queue
    void emplace(const Queue & another, int m_second, bool is_from_a, bool is_block, double coefficient); // forTail emplace
//     bool doesEmplaceMakeSense(const Queue & another, int m_second, bool is_from_a, bool is_block, double coefficient); // forTail emplace
    void mixQueues(const ABPMap<Queue>& queues, const ABPMap<double>& coefficients, const ABPMap<double>& coefficients_full, bool is_free_departure_possible);
    void mixQueues(const ABPMap<Queue>& queues, const std::map<char,double>& coefficients, const std::map<char,double>& coefficients_full, bool is_free_departure_possible);
//     Queue(const Queue & another, int new_start, int new_end); // this is effectively getMiddlePart of the another queue
    void addDepartureQueue(const Queue & another, int m_second, double coef_no_dep, double coef_dep/*, bool is_split_queue=false*/);
    Queue& operator=(const Queue& another);
    void printQueue() const;
    int getCapacity() const { return capacity; }
    double operator[](int i) const;
    void setProbability(int i, double value);
    void addProbability(int i, double value);
    Queue afterArrival(bool is_free_departure_possible) const;
//     void addArrival(bool is_free_departure_possible);
    Queue afterDeparture(int n);
    Queue afterNoDeparture(int n);
    Queue forTail(int step, int i_from, bool is_from_a, bool is_block, double coefficient);
    void removeTailPart(int new_start, int new_end, double coefficient);
    Queue getUpstreamSide(int second, int downstream_second, bool to_delete = false);
    Queue getDownstreamPart(int second, int downstream_second);
    Queue getMiddlePart(int second, int downstream_second);
    void getStartEndMiddle(int m_down, int m_up, int & new_start, int& new_end);
    void getStartEndDownstream(int m_second, int & new_start, int& new_end);
    bool removeDownstreamPart(int second, int downstream_second);
    bool removeDownstreamPart(int m_second);
    void removeMiddlePart(int second, int downstream_second);
    void removeMiddlePartKnowingMs(int m_down, int m_up, bool make_zero = true);
    
    void renormaliseQueue(double divide_by);
    
    void getStartEnd(int & get_start, int & get_end) {get_start = start; get_end = end;}
    int getStart() const { return start; }
    int getEnd() const{ return end; }
    
    void addQueue(const Queue & another);
    void addUpstreamAndMiddleQueueByCoef(const Queue& another, double coefficient, double coefficient_full, bool is_a_queue, int second, int downstream_second);
    void addDownstreamPartOfQueue(const Queue& another, int second, int downstream_second);
    double getTotalProbability() const;
//     double getPartialProbability(int from, int to) const;
//     double getPartialProbability(int from);
    bool isPartialProbabilitySignificant(int from, int to, double coef = 1.) const; // > almost_zero
//     bool isPartialProbabilitySignificant(int from, double coef = 1.); // > almost_zero
    double getAlmostFullProbability(bool is_a_queue) const;
    double getExpectedQueueLength() const;
    double getExpectedQueueSquare(int start_from) const;
    double getExpectedQueueLengthInMeters() const;
    bool hasZeroProbability() const;
    bool hasAlmostZeroProbability() const;
    
    void emptyQueue() {start = 0; end = 0;}
    
    template<typename... Args> bool doesEmplaceMakeSense(const Args &... args){ return ::doesEmplaceMakeSense(args...);}
    
    friend Queue mixQueues(const std::vector<Queue> & queues, const std::vector<double> & coefficients);
    friend Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients);
    friend Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients, 
                           const std::map<char, double> & coefficients_full);
    friend void swapQueueParts(Queue & queue1, Queue & queue2, int from, int to);
};

double stopLineSecond(int second, int m);
int minNotAffectedQueueLength(int second, int downstream_second, int capacity);

Queue mixQueues(const std::vector<Queue> & queues, const std::vector<double> & coefficients);
Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients);
Queue mixQueues(int capacity, const std::map<char, Queue> & queues, const std::map<char, double> & coefficients, 
                const std::map<char, double> & coefficients_full);

void swapQueueParts(Queue & queue1, Queue & queue2, int from, int to);


#endif // QUEUE_H
