#ifndef QUEUES_H
#define QUEUES_H
#include <vector>
#include <map>
#include "rollqueues.h"
#include "queue.h"
#include "state.h"
#include "arrivalchain.h"
#include "departurechain.h"
#include "tail.h"

//using Pair = std::pair<int, int>;

// class UntransitionedQueues;

class Queues{
    int capacity;
    std::vector<DepArrQueues> queues;  
    UntransitionedQueues untransitioned_queues;
    std::map<char, std::map<char, double>> last_ar_transitions;
    std::vector<std::map<KeyDA, double>> emptied_probs;
    
    void deleteEmptyQueues(int second);
    void inialiseInternalQueues(Pair first_d_key, int queue_length, ArrivalChain & arrival_chain, DepartureChain & departure_chain, int starting_second);
    void inialiseInternalQueues(int travel_time, Pair first_d_key);
    
    void renormaliseQueues(int second, double total);
    void checkProblemQueue(int problem_index);
    
//     void prefillEmptied(int step);
public:
    Queues(int duration, int capacity, std::map<char, ArrivalState>& a_states, Pair d_key);
    Queues(int duration, int capacity, int travel_time, int short_travel_time, Pair first_d_key);
    Queues(int duration, int capacity, int queue_length, Pair d_key, bool is_arrival, int starting_second);  // start from state queue_length
    Queues(int duration, int capacity, int travel_time, int queue_length, Pair first_d_key, 
           ArrivalChain & arrival_chain, DepartureChain & departure_chain, int starting_second);
    ~Queues();
    void printQueues(int second);
    void printUntransitionedQueues(int second) {untransitioned_queues.printQueues(second);}
    void printKeys(int second);
    void updateQueues(int step, ArrivalChain& arrival_chain, DepartureChain& departure_chain);  // makes queues for second + 1
    void updateInternalQueues(int step, int travel_time, ArrivalChain& arrival_chain, DepartureChain& departure_chain);  // makes queues for second + 1
    void updateLastSecond(int step, int travel_time, ArrivalChain& arrival_chain, DepartureChain& departure_chain);
    void updateDeparturesAndQueuesOneStep(int step, int second, DepArrQueues& new_queues, ArrivalChain& arrival_chain, DepartureChain& departure_chain);
    double getAverageQueue(int second);
    double getAverageQueueInMeters(int second);
    std::vector<double> getNormalAverages(int duration);
    bool checkAndPrintTotalProbability(int second, bool throw_exception = true, bool print = true);
    std::map<KeyDA, double> getZeroProbabilities(int second);
    
    double getAlmostFullProbability(int second, char a_key);
    double getAlmostFullUntransitionedProbability(int second, char a_key);
    std::map<KeyIA, std::map<Pair, double>> getAlmostFullDKeys(int second);
//     std::map<Pair, double> getAlmostFullDKeys(int second, char a);
//     std::map<KeyDA, double> getJustEmptiedProbabilities(int second, DepartureChain&  departure_chain, 
//                                                             std::map<char, std::map<char, double>>& arr_transitions);

    void updateDepartureState(int step, const std::map<KeyDA, std::map<Pair, double>>& tr_flows_zero);
    
    
    double processTail(int step, const Tail & tail, ArrivalChain & arrival_chain, DepartureChain & departure_chain);
    
    void addQueues(int second, DepArrQueues & new_queues);
    
    std::map<char, double> getArTransitions(int second, char a_key, ArrivalChain& arrival_chain);
    std::map<char, std::map<char, double> > getAllArTransitions(int second, ArrivalChain& arrival_chain);
    std::map<char, double> getArTransitionsFull(int second, char a_key, ArrivalChain& arrival_chain);
    std::map<char, std::map<char, double> > getAllArTransitionsFull(int second, ArrivalChain& arrival_chain);
    std::map<char, std::map<char, double> > getAllArTransitionsToKeys(int second, ArrivalChain& arrival_chain);
    std::map<char, std::map<char, double> > getAllArTransitionsFullToKeys(int second, ArrivalChain& arrival_chain);
    
    int getSize() {return queues.size();}
    
    void checkCapacities();
    bool checkProbabilities(int second, const ArrivalChain& arrival_chain);
    void checkPartialProbabilities(int second, ArrivalChain& arrival_chain);
    
    double getTotalUntransitionedQueue(int second, int downstream_second, char a_key);
    std::map<char, std::map<char, double>> getLastArrivalTransitions() {return last_ar_transitions;}
    std::map<KeyDA, double> getJustEmptiedProbabilities(int step, bool is_split_lane, bool is_internal);
    
    std::vector<Pair> cleanUnprobableStates(int step, DepartureChain & departure_chain);
    std::vector<double> getDistribution(int second);
    std::vector<double> getTiltedCumulativeDistribution(int second);
    std::vector<std::vector<double>> getRightCumulativeDistributions(int start, int end);
    
    int getTotalNumberOfQueues();
    bool checkNonZeroProbs(int step, DepartureChain & departure_chain, bool print_all = false);
    
    double getAverageAndRemove(int second);
    double getAverageSquareAndRemove(int second, int start_from);
    double getAverageSquareQueue(int second, int start_from);
    
//     friend std::pair<DepArrQueues,double> updateQueues(int second, int downstream_second, int capacity, const DepArrQueues& queues, 
//                                      std::map<char, std::map<char, double>>& ar_transitions_to, 
//                                      std::map<char, std::map<char, double>>& ar_transitions_full_to, DepartureChain& departure_chain);
//     friend double updateDepartures(int second, int downstream_second, int capacity, DepArrQueues &queues, 
//                                                           DepartureChain& departure_chain);
    
};


// double getTotalProbabilityDepArrQueues(DepArrQueues & queues);
// void printDepArrQueues(const DepArrQueues & queues);
// void deleteEmptyDepArrQueues(DepArrQueues & queues);
double printTotalProbabilityDepArrQueues(const DepArrQueues & queues);
DepArrQueues updateQueuesOneStep(int second, double & unblocked, int downstream_second, int capacity, const DepArrQueues& queues, 
                                 std::map<char, std::map<char, double>>& ar_transitions_to, 
                                 std::map<char, std::map<char, double>>& ar_transitions_full_to, 
                                 ArrivalChain & arrival_chain, DepartureChain& departure_chain, 
                                 std::vector<std::map<KeyDA, double>> & emptied);
double updateDepartures(int second, int downstream_second, int capacity, DepArrQueues &queues, DepartureChain& departure_chain, 
                        std::map<KeyDA, double> & emptied);
ABPMap<ABPMap<double>> changeArrTransitionsToABPMap(std::map<char, std::map<char, double>>& ar_transitions);
std::map<char, std::map<char, double>> flipKeys(const std::map<char, std::map<char, double>> & transitions);

#endif // QUEUES_H
