#include "state.h"

void ArrivalState::addProbability(double value){
    probability +=  value;
}


double ArrivalState::getUnknownTransitionProbability(){
//     std::cout << "getting untransitioned probability\n";
//     std::cout << probability << " (" << 1 - probability << ")\n";
//     transition_flows.printTransitions();
//     std::cout << transition_flows.getTotalTransitionProbability() << " " << 1 - transition_flows.getTotalTransitionProbability() << "\n";  
    return probability - transition_flows.getTotalTransitionProbability();
}

double ArrivalState::getUntransitionedProbability(){
    return probability - transitioned_flows.getTotalTransitionProbability() - transitioned_flows_full.getTotalTransitionProbability();
}


double ArrivalState::addDefaultBTransition(double unblocked){
    transition_flows.addTransition('p', unblocked);
    return transition_flows.addDefaultTransition('b', probability);
}

void ArrivalState::addTransitionedProbabilities(MapChar& transitioned, MapChar& transitioned_full){
    transitioned_flows.addTransitions(transitioned);
    transitioned_flows_full.addTransitions(transitioned_full);
}

bool ArrivalState::checkTransitionedFlows(bool should_be_equal){
    // transitioned should be equal to all!
    double total, prob_should_be, prob_already;
    total = 0;
    for (const auto & a_key: {'a', 'b', 'p'}){
        if (transition_flows.hasKey(a_key)){
            prob_should_be = transition_flows.getTransition(a_key);
            total += prob_should_be;
            prob_already = transitioned_flows.getTransition(a_key) + transitioned_flows_full.getTransition(a_key);
//             std::cout << total << " " << prob_should_be << " " << prob_already << "\n";
            if ((should_be_equal && !almostEqual(prob_already, prob_should_be, small_eps)) || 
                (!should_be_equal && prob_already > prob_should_be + small_eps)){
                std::cout << probability << "\n";
                std::cout << prob_already << " " << prob_should_be << " " << prob_should_be - prob_already << "\n";
                printTransitions();
                std::cout << "Transitioned more (or less) than was possible\n";
//                 throw "Transitioned more (or less) than was possible";
                return false;
            }
        }
//         std::cout << 1-total << " " <<  1- probability << "\n";
    }
    if ((should_be_equal && !almostEqual(total, probability, small_eps)) || (!should_be_equal && total > probability + small_eps)){
        std::cout << total << " " << probability << " " << total - probability << "\n";
        printTransitions();
//         throw "Together transitioned more (or less) than was possible";
        std::cout << "Together transitioned more (or less) than was possible";
        return false;
    }
    return true;
}


void ArrivalState::getTransitionsForUntransitioned(MapCharChar& transitions, char a_key_from, double almost_full, bool is_fully_known){
    double total_not_transitioned = probability - transitioned_flows.getTotalTransitionProbability() 
                                    - almost_full;
//     std::cout << almost_full << " " << probability << " " << transitioned_flows.getTotalTransitionProbability() << "\n";
    if (total_not_transitioned > almost_zero){
        if (is_fully_known){
            double to_a = std::max(transition_flows.getTransition('a') - transitioned_flows.getTransition('a'), 0.);
//             std::cout << total_not_transitioned << " " << to_a << " <- trying to find transitions\n";
//             std::cout << to_a - total_not_transitioned << "\n";
//             std::cout << almost_full << " " << probability << " " << transitioned_flows.getTotalTransitionProbability() 
//                       << " " << transition_flows.getTransition('a') << " " << transitioned_flows.getTransition('a') << "\n";
//             transitioned_flows.printTransitions();
                      
            if (to_a > 0){
                double prob = to_a / total_not_transitioned;
                transitions['a'][a_key_from] = std::min(prob, 1.);
                if (to_a < total_not_transitioned)
                    transitions['p'][a_key_from] = std::min(1 - prob, 1.);
            } else {
                transitions['p'][a_key_from] = 1;
            }
        } else {
//             if (probability > almost_zero + almost_full){
//                 transitions['a'][a_key_from] = transition_flows.getTransition('a') / (probability - almost_full);
//                 transitions['p'][a_key_from] = transition_flows.getTransition('p') / (probability - almost_full);
//             }
            // in this case not all transitions are known, but full states only transition to state 'b' if any
            double to_a = transition_flows.getTransition('a') - transitioned_flows.getTransition('a');
            if (to_a > almost_zero)
                transitions['a'][a_key_from] = to_a / total_not_transitioned;
            double to_p = transition_flows.getTransition('p') - transitioned_flows.getTransition('p');
            if (to_p > almost_zero)
                transitions['p'][a_key_from] = to_p / total_not_transitioned;
        }
    }
}


void ArrivalState::getTransitionsForUntransitioned(MapCharChar& transitions, char a_key_from, double almost_full, double partially_not_transitioned){
    double total_not_transitioned = probability - transitioned_flows.getTotalTransitionProbability() 
                                    - almost_full - partially_not_transitioned;
    if (total_not_transitioned > almost_zero){
        double to_a = transition_flows.getTransition('a') - transitioned_flows.getTransition('a');
        if (to_a > almost_zero)
            transitions['a'][a_key_from] = to_a / total_not_transitioned;
        double to_p = transition_flows.getTransition('p') - transitioned_flows.getTransition('p');
        if (to_p > almost_zero)
            transitions['p'][a_key_from] = to_p / total_not_transitioned;
    }
}


void ArrivalState::getTransitionsForUntransitionedFull(MapCharChar& transitions, char a_key_from, double almost_full, bool is_fully_known){
    double total_not_transitioned = almost_full - transitioned_flows_full.getTotalTransitionProbability();
    if (total_not_transitioned > almost_zero){
        if (is_fully_known){
            double to_b = std::max(0., transition_flows.getTransition('b') - transitioned_flows_full.getTransition('b'));
            if (to_b > 0){
                double prob = to_b / total_not_transitioned;
                transitions['b'][a_key_from] = std::min(prob, 1.);
                if (to_b + small_eps < total_not_transitioned)
                    transitions['p'][a_key_from] = std::min(1 - prob, 1.);
            } else {
                transitions['p'][a_key_from] = 1;
            }
        } else {  // this transitions shouldn't be used
//             std::cout << "in no transition part?\n";
            // in this case not all transitions are known, but only state 'a' may have transitions and only to 'b' if any
//             if (a_key_from == 'a'){
//                 double to_b = transition_flows.getTransition('b') - transitioned_flows_full.getTransition('b');
//                 if (to_b > small_eps)
//                     transitions['b'][a_key_from] = to_b / total_not_transitioned;
//             }
        }
    }
}

void ArrivalState::printTransitions() { 
    std::cout << "transition flows:\n";
    transition_flows.printTransitions(); 
    std::cout << "transitioned flows:\n";
    transitioned_flows.printTransitions();
    std::cout << "transitioned flows full:\n";
    transitioned_flows_full.printTransitions();
}


void DepartureState::addTransition(Pair pair, double probability, double departure_probability){
    transitions.addTransition(pair, probability);
    departure_probabilities.addTransition(pair, departure_probability);
}

void DepartureState::printState(Pair d_key){
    std::cout << "\t" << d_key << ": last second " << getLastSecond() << ", transition probabilities:\n";
    transitions.printTransitions();
    std::cout << "departure probabilities during transitions:\n";
    departure_probabilities.printTransitions();
    std::cout << "probability: " << getProbability() << "\n";
}

bool DepartureState::removeProbability(Pair default_transition, double prob){
    // update transition probabilities since probability_was
    if (probability < prob) {
        if (probability + small_eps < prob){
            std::cout << "Have not removed " << prob << " from " << probability << ": it is too much!\n";
            return false;
        } else {
            prob = probability;
        }
    }
    transitions.rebalanceTransitions(default_transition, probability, prob);
    probability -= prob;
    return true;
}

