#ifndef NODE_H
#define NODE_H
#include <map>
#include "transitions.h"
#include "general.h"


class ArrivalState{ //: public Node<int>{
    double probability;
    Transitions<char> transition_flows;
    Transitions<char> transitioned_flows;
    Transitions<char> transitioned_flows_full;
public:
    ArrivalState(): probability(0.) {}
    explicit ArrivalState (double probability): probability(probability) {}
    void addProbability(double value);
    double getProbability() const { return probability;}
    double getUnknownTransitionProbability();
    double getUntransitionedProbability();
    
    double getTransition(char a_key) { return transition_flows.getTransition(a_key); }
    void addTransition(char a_key, double value) { transition_flows.addTransition(a_key, value); }
    void printTransitions();
    std::map< char, double > getNormalisedTransitions(double state_probability) { return transition_flows.getNormalisedTransitions(state_probability); }
    bool checkTransitions(double total) { return transition_flows.checkTransitions(total); }
    double addDefaultTransition(char a_key_to) { return transition_flows.addDefaultTransition(a_key_to, probability); }
    double addDefaultBTransition(double unblocked);
    
    void addTransitionedProbabilities(MapChar & transitioned, MapChar & transitioned_full);
    void getTransitionsForUntransitioned(MapCharChar & transitions, char a_key_from, double almost_full, bool is_fully_known);
    void getTransitionsForUntransitioned(MapCharChar & transitions, char a_key_from, double almost_full, double partially_not_transitioned);
    void getTransitionsForUntransitionedFull(MapCharChar & transitions, char a_key_from, double almost_full, bool is_fully_known);
    bool checkTransitionedFlows(bool should_be_equal);
};

std::ostream& operator<<(std::ostream& os, const Pair & pair);

class DepartureState  { //: public Node<int>{
    int last;
    double probability;
//     bool is_green_begin;
public:
    Transitions<Pair> transitions;
    Transitions<Pair> departure_probabilities;
    DepartureState() : last(-1), probability(-1) { throw "Default departure state constructor is called!\n"; }
    explicit DepartureState(bool dummy) : last(-1), probability(0) { if (!dummy) {throw "Normal departure state constructor is called with wrong argument!\n";}}
//     DepartureState(bool is_green_begin): is_green_begin(is_green_begin) { probability = 0; }
    void updateLastSecond(int value){ last = value; }
    double getLastSecond(){ return last; }
    void addTransition(Pair pair, double probability, double departure_probability);
    
    void addProbability(double value) { probability += value; }
    double getProbability(){ return probability;}
    
    void printState(Pair d_key);
    
    bool removeProbability(Pair default_transition, double prob);
    
//     bool if_green_begin() {return is_green_begin;}
};

#endif // NODE_H
